"""
Helper script to create a new plugin structure
"""

import os
import sys

import yaml

# current directory:
current_dir = os.path.dirname(os.path.realpath(__file__))


class PluginStructure:
    """
    Example usage. Run this script with the target directory as an argument
    It will then automatically create the plugin structure in the target directory

    >>> python create_plugin_structure.py /path/to/my_plugin
    """

    def __init__(self, target_dir):
        """This class can be used to produce the folder structure
        of BEC. This includes copying templates for the structures
        """
        self.target_dir = target_dir
        _, self.plugin_name = os.path.split(target_dir)
        self.create_dir("")

    def create_dir(self, dir_name):
        os.makedirs(os.path.join(self.target_dir, dir_name), exist_ok=True)

    def create_init_file(self, dir_name):
        init_file = os.path.join(self.target_dir, dir_name, "__init__.py")
        with open(init_file, "w", encoding="utf-8") as f:
            f.write("")

    def copy_plugin_setup_files(self):
        # copy setup files
        self.copy_toml_file()
        git_hooks = os.path.join(current_dir, "plugin_setup_files", "git_hooks/*")
        self.create_dir(".git_hooks")
        os.system(f"cp -R {git_hooks} {self.target_dir}/.git_hooks/")
        gitignore = os.path.join(current_dir, "plugin_setup_files", "gitignore")
        os.system(f"cp {gitignore} {self.target_dir}/.gitignore")

    def copy_toml_file(self):
        """Copy the toml file and change the template name in the file"""
        # copy toml file
        toml_file = os.path.join(current_dir, "plugin_setup_files", "pyproject.toml")
        with open(toml_file, "r", encoding="utf-8") as f:
            toml_template_str = f.read()
        # change template name in toml file
        toml_template_str = toml_template_str.replace("{template_name}", self.plugin_name)

        toml_file = os.path.join(self.target_dir, "pyproject.toml")
        # write toml file
        with open(toml_file, "w", encoding="utf-8") as f:
            f.write(toml_template_str)

    def add_plugins(self):
        self.create_dir(self.plugin_name)
        self.create_init_file(self.plugin_name)

    def add_scans(self):
        self.create_dir(f"{self.plugin_name}/scans")
        self.create_init_file(f"{self.plugin_name}/scans")

        # copy scan_plugin_template.py
        scan_plugin_template_file = os.path.join(
            current_dir, "plugin_setup_files", "scan_plugin_template.py"
        )
        os.system(f"cp {scan_plugin_template_file} {self.target_dir}/{self.plugin_name}/scans")

    def add_client(self):
        self.create_dir(f"{self.plugin_name}/bec_ipython_client")
        self.create_init_file(f"{self.plugin_name}/bec_ipython_client")

        # high level interface
        self.create_dir(f"{self.plugin_name}/bec_ipython_client/high_level_interface")
        self.create_init_file(f"{self.plugin_name}/bec_ipython_client/high_level_interface")

        # plugins
        self.create_dir(f"{self.plugin_name}/bec_ipython_client/plugins")
        self.create_init_file(f"{self.plugin_name}/bec_ipython_client/plugins")

        # startup
        self.create_dir(f"{self.plugin_name}/bec_ipython_client/startup")
        self.create_init_file(f"{self.plugin_name}/bec_ipython_client/startup")

        ## copy pre_startup.py
        pre_startup_file = os.path.join(current_dir, "plugin_setup_files", "pre_startup.py")
        os.system(
            f"cp {pre_startup_file} {self.target_dir}/{self.plugin_name}/bec_ipython_client/startup"
        )
        ## copy post_startup.py
        post_startup_file = os.path.join(current_dir, "plugin_setup_files", "post_startup.py")
        os.system(
            f"cp {post_startup_file} {self.target_dir}/{self.plugin_name}/bec_ipython_client/startup"
        )

    def add_devices(self):
        self.create_dir(f"{self.plugin_name}/devices")
        self.create_init_file(f"{self.plugin_name}/devices")
        # device template?

    def add_device_configs(self):
        self.create_dir(f"{self.plugin_name}/device_configs")
        self.create_init_file(f"{self.plugin_name}/device_configs")

    def add_services(self):
        self.create_dir(f"{self.plugin_name}/services")
        self.create_init_file(f"{self.plugin_name}/services")

    def add_bec_widgets(self):
        self.create_dir(f"{self.plugin_name}/bec_widgets")
        self.create_init_file(f"{self.plugin_name}/bec_widgets")

        self.create_dir(f"{self.plugin_name}/bec_widgets/configs")
        self.create_dir(f"{self.plugin_name}/bec_widgets/widgets")
        self.create_init_file(f"{self.plugin_name}/bec_widgets/widgets")

    def add_file_writer(self):
        self.create_dir(f"{self.plugin_name}/file_writer")
        self.create_init_file(f"{self.plugin_name}/file_writer")

    def add_deployments(self):
        self.create_dir(f"{self.plugin_name}/deployments")
        self.create_init_file(f"{self.plugin_name}/deployments")
        self.create_dir(f"{self.plugin_name}/deployments/device_server")
        self.create_init_file(f"{self.plugin_name}/deployments/device_server")

        ds_startup = os.path.join(current_dir, "plugin_setup_files", "setup_device_server.py")
        os.system(
            f"cp {ds_startup} {self.target_dir}/{self.plugin_name}/deployments/device_server/startup.py"
        )

    def add_gitlab_ci(self):
        out = {
            "include": [
                {
                    "project": "bec/awi_utils",
                    "file": "/templates/plugin-repo-template.yml",
                    "inputs": {"name": self.plugin_name, "target": self.plugin_name},
                }
            ]
        }
        with open(os.path.join(self.target_dir, ".gitlab-ci.yml"), "w", encoding="utf-8") as f:
            yaml.dump(out, f)

    def add_tests(self):
        self.create_dir("tests/tests_bec_ipython_client")
        self.copy_tests_readme("tests/tests_bec_ipython_client")
        self.create_dir("tests/tests_dap_services")
        self.copy_tests_readme("tests/tests_services")
        self.create_dir("tests/tests_bec_widgets")
        self.copy_tests_readme("tests/tests_bec_widgets")
        self.create_dir("tests/tests_devices")
        self.copy_tests_readme("tests/tests_devices")
        self.create_dir("tests/tests_scans")
        self.copy_tests_readme("tests/tests_scans")
        self.create_dir("tests/tests_file_writer")
        self.copy_tests_readme("tests/tests_file_writer")

    def copy_tests_readme(self, file_path: str):
        readme_file = os.path.join(current_dir, "plugin_setup_files", "README_template_tests.md")
        os.system(f"cp {readme_file} {self.target_dir}/{file_path}/README.md")

    def add_bin(self):
        self.create_dir("bin")


if __name__ == "__main__":
    struc = PluginStructure(sys.argv[1])
    struc.add_plugins()
    struc.copy_plugin_setup_files()
    struc.add_scans()
    struc.add_client()
    struc.add_devices()
    struc.add_device_configs()
    struc.add_services()
    struc.add_bec_widgets()
    struc.add_file_writer()
    struc.add_tests()
    struc.add_bin()
    struc.add_deployments()
    struc.add_gitlab_ci()

    print(f"Plugin structure created in {sys.argv[1]}")



# pylint: skip-file
import os
from unittest import mock

import pytest
import yaml
from fastjsonschema import JsonSchemaException
from test_scibec_connector import SciBecMock, SciHubMock

import bec_lib
from bec_lib import DeviceBase, messages
from bec_lib.bec_errors import DeviceConfigError
from bec_lib.device import OnFailure, ReadoutPriority
from bec_server.scihub import SciHub
from bec_server.scihub.scibec import ConfigHandler, SciBecConnector

dir_path = os.path.dirname(bec_lib.__file__)


@pytest.fixture()
def config_handler(SciHubMock):
    with mock.patch("bec_server.scihub.scibec.scibec_connector.os.path.exists", return_value=False):
        scibec_connector = SciBecConnector(SciHubMock, SciHubMock.connector)
    with mock.patch.object(scibec_connector, "_start_config_request_handler"):
        with mock.patch.object(scibec_connector, "_start_metadata_handler"):
            with mock.patch.object(scibec_connector, "_start_scibec_account_update"):
                scibec_connector.scibec = None
                yield scibec_connector.config_handler
                scibec_connector.shutdown()


@pytest.fixture
def available_keys_fixture(test_config_yaml):
    config_content = test_config_yaml
    config_keys_from_config = []
    for dev_config in config_content.values():
        for k in dev_config.keys():
            config_keys_from_config.append(k)
    config_keys_from_config = set(config_keys_from_config)
    # remove enabled and deviceClass, as it is not handled from config_handler
    config_keys_from_config.remove("enabled")
    config_keys_from_config.remove("deviceClass")
    return config_keys_from_config


def test_parse_config_request_update(config_handler):
    msg = messages.DeviceConfigMessage(
        action="update", config={"samx": {"enabled": True}}, metadata={}
    )
    with mock.patch.object(config_handler, "_update_config") as update_config, mock.patch.object(
        config_handler.device_manager, "check_request_validity"
    ) as req_validity:
        config_handler.parse_config_request(msg)
        req_validity.assert_called_once_with(msg)
        update_config.assert_called_once_with(msg)


def test_parse_config_request_reload(config_handler):
    msg = messages.DeviceConfigMessage(action="reload", config={}, metadata={})
    with mock.patch.object(config_handler, "_reload_config") as reload_config, mock.patch.object(
        config_handler.device_manager, "check_request_validity"
    ) as req_validity:
        config_handler.parse_config_request(msg)
        req_validity.assert_called_once_with(msg)
        reload_config.assert_called_once_with(msg)


def test_parse_config_request_set(config_handler):
    msg = messages.DeviceConfigMessage(
        action="set", config={"samx": {"enabled": True}}, metadata={}
    )
    with mock.patch.object(config_handler, "_set_config") as set_config, mock.patch.object(
        config_handler.device_manager, "check_request_validity"
    ) as req_validity:
        config_handler.parse_config_request(msg)
        req_validity.assert_called_once_with(msg)
        set_config.assert_called_once_with(msg)


def test_parse_config_request_exception(config_handler):
    msg = messages.DeviceConfigMessage(
        action="update", config={"samx": {"enabled": True}}, metadata={}
    )
    with mock.patch.object(config_handler, "send_config_request_reply") as req_reply:
        with mock.patch("bec_server.scihub.scibec.config_handler.traceback.format_exc") as exc:
            with mock.patch.object(config_handler, "_update_config", side_effect=AttributeError()):
                config_handler.parse_config_request(msg)
                req_reply.assert_called_once_with(accepted=False, error_msg=exc(), metadata={})


def test_config_handler_reload_config(config_handler):
    msg = messages.DeviceConfigMessage(action="reload", config={}, metadata={})
    with mock.patch.object(config_handler, "send_config_request_reply") as req_reply:
        with mock.patch.object(config_handler, "send_config") as send:
            config_handler.parse_config_request(msg)
            send.assert_called_once_with(msg)


### Commented out as config updates on scibec are not supported yet

# def test_config_handler_reload_config_with_scibec(SciHubMock):
#     scibec_connector = SciBecMock
#     config_handler = scibec_connector.config_handler
#     msg = messages.DeviceConfigMessage(action="reload", config={}, metadata={})
#     with mock.patch.object(config_handler, "send_config_request_reply") as req_reply:
#         with mock.patch.object(scibec_connector, "scibec"):
#             with mock.patch.object(scibec_connector, "update_session") as update_session:
#                 with mock.patch.object(config_handler, "send_config") as send:
#                     config_handler.parse_config_request(msg)
#                     send.assert_called_once_with(msg)
#                     update_session.assert_called_once()


@pytest.mark.parametrize(
    "config, expected",
    [
        ({"samx": {"enabled": True}}, {"name": "samx", "enabled": True}),
        (
            {"samx": {"enabled": True, "deviceConfig": None}},
            {"name": "samx", "enabled": True, "deviceConfig": {}},
        ),
    ],
)
def test_config_handler_set_config(config_handler, config, expected):
    msg = messages.DeviceConfigMessage(action="set", config=config, metadata={"RID": "12345"})
    with mock.patch.object(config_handler.validator, "validate_device") as validator:
        with mock.patch.object(config_handler, "send_config_request_reply") as req_reply:
            with mock.patch.object(
                config_handler,
                "_wait_for_device_server_update",
                return_value=(True, mock.MagicMock()),
            ) as wait:
                with mock.patch.object(config_handler, "send_config") as send_config:
                    config_handler._set_config(msg)
                    req_reply.assert_called_once_with(
                        accepted=True,
                        error_msg=None,
                        metadata={"RID": "12345", "updated_config": True},
                    )
                    validator.assert_called_once_with(expected)
                    send_config.assert_called_once_with(
                        messages.DeviceConfigMessage(
                            action="reload",
                            config={},
                            metadata={"RID": "12345", "updated_config": True},
                        )
                    )


def test_config_handler_set_invalid_config_raises(config_handler):
    msg = messages.DeviceConfigMessage(
        action="set", config={"samx": {"status": {"enabled": True}}}, metadata={"RID": "12345"}
    )
    with pytest.raises(JsonSchemaException):
        with mock.patch.object(config_handler, "send_config_request_reply") as req_reply:
            config_handler._set_config(msg)
            req_reply.assert_called_once_with(
                accepted=True, error_msg=None, metadata={"RID": "12345"}
            )


### Commented out as config updates on scibec are not supported yet

# def test_config_handler_set_config_with_scibec(SciHubMock):
#     scibec_connector = SciBecMock
#     config_handler = scibec_connector.config_handler
#     msg = messages.DeviceConfigMessage(
#         action="set", config={"samx": {"enabled": True}}, metadata={}
#     )
#     scibec_connector.scibec_info = {"beamline": {"info": [], "activeExperiment": "12345"}}
#     with mock.patch.object(scibec_connector, "scibec") as scibec:
#         with mock.patch.object(config_handler, "send_config_request_reply") as req_reply:
#             with mock.patch.object(scibec_connector, "update_session") as update_session:
#                 config_handler._set_config(msg)
#                 scibec.set_session_data.assert_called_once_with(
#                     "12345", {"samx": {"enabled": True}}
#                 )
#                 req_reply.assert_called_once_with(accepted=True, error_msg=None, metadata={})
#                 update_session.assert_called_once()


def test_config_handler_update_config(config_handler):
    dev = config_handler.device_manager.devices
    dev.samx = DeviceBase(name="samx", config={})
    msg = messages.DeviceConfigMessage(
        action="update", config={"samx": {"enabled": True}}, metadata={}
    )
    with mock.patch.object(
        config_handler, "_update_device_config", return_value=True
    ) as update_device_config:
        with mock.patch.object(config_handler, "update_config_in_redis") as update_config_in_redis:
            with mock.patch.object(config_handler, "send_config") as send_config:
                with mock.patch.object(
                    config_handler, "send_config_request_reply"
                ) as send_config_request_reply:
                    config_handler._update_config(msg)
                    update_device_config.assert_called_once_with(dev["samx"], {"enabled": True})
                    update_config_in_redis.assert_called_once_with(dev["samx"])

                    send_config.assert_called_once_with(msg)
                    send_config_request_reply.assert_called_once_with(
                        accepted=True, error_msg=None, metadata={}
                    )


def test_config_handler_update_config_not_updated(config_handler):
    dev = config_handler.device_manager.devices
    dev.samx = DeviceBase(name="samx", config={})
    msg = messages.DeviceConfigMessage(
        action="update", config={"samx": {"enabled": True}}, metadata={}
    )
    with mock.patch.object(
        config_handler, "_update_device_config", return_value=False
    ) as update_device_config:
        with mock.patch.object(config_handler, "update_config_in_redis") as update_config_in_redis:
            with mock.patch.object(config_handler, "send_config") as send_config:
                with mock.patch.object(
                    config_handler, "send_config_request_reply"
                ) as send_config_request_reply:
                    config_handler._update_config(msg)
                    update_device_config.assert_called_once_with(dev["samx"], {"enabled": True})
                    update_config_in_redis.assert_not_called()

                    send_config.assert_not_called()
                    send_config_request_reply.assert_not_called()


def test_config_handler_update_device_config_enable(config_handler):
    dev = config_handler.device_manager.devices
    dev.samx = DeviceBase(name="samx", config={})
    with mock.patch.object(config_handler, "_update_device_server") as update_dev_server:
        with mock.patch.object(
            config_handler, "_wait_for_device_server_update", return_value=(True, mock.MagicMock())
        ) as wait:
            with mock.patch("bec_server.scihub.scibec.config_handler.uuid") as uuid:
                device = dev["samx"]
                rid = str(uuid.uuid4())
                config_handler._update_device_config(device, {"enabled": True})
                # mock doesn't copy the data, hence the popped result:
                update_dev_server.assert_called_once_with(rid, {device.name: {}})
                wait.assert_called_once_with(rid)


def test_config_handler_update_device_config_deviceConfig(config_handler):
    dev = config_handler.device_manager.devices
    dev.samx = DeviceBase(name="samx", config={"deviceConfig": {}})
    with mock.patch.object(config_handler, "_update_device_server") as update_dev_server:
        with mock.patch.object(
            config_handler, "_wait_for_device_server_update", return_value=(True, mock.MagicMock())
        ) as wait:
            with mock.patch("bec_server.scihub.scibec.config_handler.uuid") as uuid:
                device = dev["samx"]
                rid = str(uuid.uuid4())
                config_handler._update_device_config(
                    device, {"deviceConfig": {"something": "to_update"}}
                )
                # mock doesn't copy the data, hence the popped result:
                update_dev_server.assert_called_once_with(rid, {device.name: {}})
                wait.assert_called_once_with(rid)
                assert dev.samx._config == {"deviceConfig": {"something": "to_update"}}


def test_config_handler_update_device_config_misc(config_handler):
    dev = config_handler.device_manager.devices
    dev.samx = DeviceBase(name="samx", config={})
    with mock.patch.object(config_handler, "_validate_update") as validate_update:
        device = dev["samx"]
        config_handler._update_device_config(device, {"readOnly": True})
        validate_update.assert_called_once_with({"readOnly": True})


def test_config_handler_update_device_config_raise(config_handler):
    dev = config_handler.device_manager.devices
    dev.samx = DeviceBase(name="samx", config={})
    with mock.patch.object(config_handler, "_validate_update") as validate_update:
        device = dev["samx"]
        with pytest.raises(DeviceConfigError):
            config_handler._update_device_config(device, {"doesnt_exist": False})


def test_config_handler_update_device_config_available_keys(config_handler, available_keys_fixture):
    for available_key in available_keys_fixture:
        dev = config_handler.device_manager.devices
        if available_key in ["deviceConfig", "userParameter"]:
            init = {"something": "to_update"}
            dev.samx = DeviceBase(name="samx", config={available_key: init})
        elif available_key in ["softwareTrigger", "readOnly"]:
            init = True
            dev.samx = DeviceBase(name="samx", config={available_key: init})
        elif available_key in ["readoutPriority"]:
            init = ReadoutPriority.BASELINE
            dev.samx = DeviceBase(name="samx", config={available_key: init})
        elif available_key in ["onFailure"]:
            init = OnFailure.BUFFER
            dev.samx = DeviceBase(name="samx", config={available_key: init})
        elif available_key in ["deviceTags"]:
            init = ["something"]
            dev.samx = DeviceBase(name="samx", config={available_key: init})
        else:
            dev.samx = DeviceBase(name="samx", config={})
        with mock.patch.object(config_handler, "_update_device_server") as update_dev_server:
            with mock.patch.object(
                config_handler,
                "_wait_for_device_server_update",
                return_value=(True, mock.MagicMock()),
            ) as wait:
                with mock.patch("bec_server.scihub.scibec.config_handler.uuid") as uuid:
                    device = dev["samx"]
                    rid = str(uuid.uuid4())
                    if available_key in ["deviceConfig", "userParameter"]:
                        update = {"something": "to_update"}
                        config_handler._update_device_config(device, {available_key: update})
                    elif available_key in ["softwareTrigger", "readOnly"]:
                        update = True
                        config_handler._update_device_config(device, {available_key: update})
                    elif available_key in ["readoutPriority"]:
                        update = ReadoutPriority.MONITORED
                        config_handler._update_device_config(device, {available_key: update})
                    elif available_key in ["readoutPriority"]:
                        update = OnFailure.RETRY
                        config_handler._update_device_config(device, {available_key: update})
                    elif available_key in ["deviceTags"]:
                        update = ["something"]
                        config_handler._update_device_config(device, {available_key: update})
                    else:
                        update = ""
                        config_handler._update_device_config(device, {available_key: update})
                    # mock doesn't copy the data, hence the popped result:
                    if available_key == "deviceConfig":
                        update_dev_server.assert_called_once_with(rid, {device.name: {}})
                        wait.assert_called_once_with(rid)
                    assert dev.samx._config == {available_key: update}


def test_config_handler_wait_for_device_server_update(config_handler):
    RID = "12345"
    with mock.patch.object(config_handler.connector, "get") as mock_get:
        mock_get.side_effect = [
            None,
            None,
            None,
            messages.RequestResponseMessage(accepted=True, message=""),
        ]
        config_handler._wait_for_device_server_update(RID)


def test_config_handler_wait_for_device_server_update_timeout(config_handler):
    RID = "12345"
    with mock.patch.object(config_handler.connector, "get", return_value=None) as mock_get:
        with pytest.raises(TimeoutError):
            config_handler._wait_for_device_server_update(RID, timeout_time=0.1)
            mock_get.assert_called()



"""
This module provides a class that provides a convenient way to interact with the scan queue as well
as the requests and scans that are currently running or have been completed.
"""

from __future__ import annotations

import uuid

from typeguard import typechecked

from bec_lib import messages
from bec_lib.endpoints import MessageEndpoints
from bec_lib.logger import bec_logger
from bec_lib.queue_items import QueueStorage
from bec_lib.request_items import RequestStorage
from bec_lib.scan_items import ScanStorage

logger = bec_logger.logger


class ScanManager:
    def __init__(self, connector):
        """
        ScanManager is a class that provides a convenient way to interact with the scan queue as well
        as the requests and scans that are currently running or have been completed.
        It also contains storage container for the queue, requests and scans.

        Args:
            connector (BECConnector): BECConnector instance
        """
        self.connector = connector
        self.queue_storage = QueueStorage(scan_manager=self)
        self.request_storage = RequestStorage(scan_manager=self)
        self.scan_storage = ScanStorage(scan_manager=self)

        self.connector.register(
            topics=MessageEndpoints.scan_queue_status(), cb=self._scan_queue_status_callback
        )
        self.connector.register(
            topics=MessageEndpoints.scan_queue_request(), cb=self._scan_queue_request_callback
        )
        self.connector.register(
            topics=MessageEndpoints.scan_queue_request_response(),
            cb=self._scan_queue_request_response_callback,
        )
        self.connector.register(
            topics=MessageEndpoints.scan_status(), cb=self._scan_status_callback
        )

        self.connector.register(
            topics=MessageEndpoints.scan_segment(), cb=self._scan_segment_callback
        )

        self.connector.register(topics=MessageEndpoints.scan_baseline(), cb=self._baseline_callback)

        self.connector.register(topics=MessageEndpoints.client_info(), cb=self._client_msg_callback)

    def update_with_queue_status(self, queue: messages.ScanQueueStatusMessage) -> None:
        """update storage with a new queue status message"""
        self.queue_storage.update_with_status(queue)
        self.scan_storage.update_with_queue_status(queue)

    def request_scan_interruption(self, deferred_pause=True, scan_id: str = None) -> None:
        """request a scan interruption

        Args:
            deferred_pause (bool, optional): Request a deferred pause. If False, a pause will be requested. Defaults to True.
            scan_id (str, optional): ScanID. Defaults to None.

        """
        if scan_id is None:
            scan_id = self.scan_storage.current_scan_id
        if not any(scan_id):
            return self.request_scan_abortion()

        action = "deferred_pause" if deferred_pause else "pause"
        logger.info(f"Requesting {action}")
        return self.connector.send(
            MessageEndpoints.scan_queue_modification_request(),
            messages.ScanQueueModificationMessage(scan_id=scan_id, action=action, parameter={}),
        )

    def request_scan_abortion(self, scan_id=None):
        """request a scan abortion

        Args:
            scan_id (str, optional): ScanID. Defaults to None.

        """
        if scan_id is None:
            scan_id = self.scan_storage.current_scan_id
        logger.info("Requesting scan abortion")
        self.connector.send(
            MessageEndpoints.scan_queue_modification_request(),
            messages.ScanQueueModificationMessage(scan_id=scan_id, action="abort", parameter={}),
        )

    def request_scan_halt(self, scan_id=None):
        """request a scan halt

        Args:
            scan_id (str, optional): ScanID. Defaults to None.

        """
        if scan_id is None:
            scan_id = self.scan_storage.current_scan_id
        logger.info("Requesting scan halt")
        self.connector.send(
            MessageEndpoints.scan_queue_modification_request(),
            messages.ScanQueueModificationMessage(scan_id=scan_id, action="halt", parameter={}),
        )

    def request_scan_continuation(self, scan_id=None):
        """request a scan continuation

        Args:
            scan_id (str, optional): ScanID. Defaults to None.

        """
        if scan_id is None:
            scan_id = self.scan_storage.current_scan_id
        logger.info("Requesting scan continuation")
        self.connector.send(
            MessageEndpoints.scan_queue_modification_request(),
            messages.ScanQueueModificationMessage(scan_id=scan_id, action="continue", parameter={}),
        )

    def request_queue_reset(self):
        """request a scan queue reset"""
        logger.info("Requesting a queue reset")
        self.connector.send(
            MessageEndpoints.scan_queue_modification_request(),
            messages.ScanQueueModificationMessage(scan_id=None, action="clear", parameter={}),
        )

    def request_scan_restart(self, scan_id=None, requestID=None, replace=True) -> str:
        """request to restart a scan"""
        if scan_id is None:
            scan_id = self.scan_storage.current_scan_id
        if requestID is None:
            requestID = str(uuid.uuid4())
        logger.info("Requesting to abort and repeat a scan")
        position = "replace" if replace else "append"

        self.connector.send(
            MessageEndpoints.scan_queue_modification_request(),
            messages.ScanQueueModificationMessage(
                scan_id=scan_id,
                action="restart",
                parameter={"position": position, "RID": requestID},
            ),
        )
        return requestID

    @property
    def next_scan_number(self):
        """get the next scan number from redis"""
        msg = self.connector.get(MessageEndpoints.scan_number())
        if msg is None:
            logger.warning("Failed to retrieve scan number from redis.")
            return -1
        if not isinstance(msg, messages.VariableMessage):
            # this is a temporary fix for providing backwards compatibility
            return int(msg)
        return int(msg.value)

    @next_scan_number.setter
    @typechecked
    def next_scan_number(self, val: int):
        """set the next scan number in redis"""
        msg = messages.VariableMessage(value=val)
        return self.connector.set(MessageEndpoints.scan_number(), msg)

    @property
    def next_dataset_number(self):
        """get the next dataset number from redis"""
        msg = self.connector.get(MessageEndpoints.dataset_number())
        if msg is None:
            logger.warning("Failed to retrieve dataset number from redis.")
            return -1
        if not isinstance(msg, messages.VariableMessage):
            # this is a temporary fix for providing backwards compatibility
            return int(msg)
        return int(msg.value)

    @next_dataset_number.setter
    @typechecked
    def next_dataset_number(self, val: int):
        """set the next dataset number in redis"""
        msg = messages.VariableMessage(value=val)
        return self.connector.set(MessageEndpoints.dataset_number(), msg)

    def _scan_queue_status_callback(self, msg, **_kwargs) -> None:
        queue_status = msg.value
        if not queue_status:
            return
        self.update_with_queue_status(queue_status)

    def _scan_queue_request_callback(self, msg, **_kwargs) -> None:
        request = msg.value
        self.request_storage.update_with_request(request)

    def _scan_queue_request_response_callback(self, msg, **_kwargs) -> None:
        response = msg.value
        logger.debug(response)
        self.request_storage.update_with_response(response)

    def _client_msg_callback(self, msg: dict, **_kwargs) -> None:
        message = msg["data"]
        self.request_storage.update_with_client_message(message)

    def _scan_status_callback(self, msg, **_kwargs) -> None:
        scan = msg.value
        self.scan_storage.update_with_scan_status(scan)

    def _scan_segment_callback(self, msg, **_kwargs) -> None:
        scan_msgs = msg.value
        if not isinstance(scan_msgs, list):
            scan_msgs = [scan_msgs]
        for scan_msg in scan_msgs:
            self.scan_storage.add_scan_segment(scan_msg)

    def _baseline_callback(self, msg, **_kwargs) -> None:
        msg = msg.value
        self.scan_storage.add_scan_baseline(msg)

    def __str__(self) -> str:
        try:
            return "\n".join(self.queue_storage.describe_queue())
        except Exception:
            # queue_storage.describe_queue() can fail,
            # for example if there is no current scan queue (None)
            return super().__str__()

    def shutdown(self):
        pass



import os
import sys

import bec_lib

from ophyd_devices.utils.static_device_test import launch


def test_static_device_test():
    config_path = os.path.join(os.path.dirname(bec_lib.__file__), "configs", "demo_config.yaml")
    sys.argv = ["", "--config", config_path, "--connect"]
    launch()



import inspect
import sys
from importlib import import_module
from typing import Any

from bec_lib.utils.proxy import Proxy


def lazy_import(module_name):
    return Proxy(lambda: import_module(module_name), init_once=True)


def lazy_import_from(module_name, from_list):
    ret = (Proxy(lambda: getattr(import_module(module_name), name)) for name in from_list)
    if len(from_list) == 1:
        return next(ret)
    else:
        return ret


def isinstance_based_on_class_name(obj: Any, full_class_name: str):
    """Return if object 'obj' is an instance of class named 'full_class_name'

    'full_class_name' must be a string like 'class_module.class_name', the corresponding class does not need to be imported at the caller module level
    """
    return full_class_name in [
        f"{klass.__module__}.{klass.__name__}" for klass in inspect.getmro(type(obj))
    ]



# Description: Launch the SciHub connector.
# This script is the entry point for the SciHub connector. It is called either
# by the bec-dap entry point or directly from the command line.
import argparse
import threading

from bec_lib import RedisConnector, ServiceConfig, bec_logger
from bec_server import scihub

logger = bec_logger.logger
bec_logger.level = bec_logger.LOGLEVEL.INFO


def main():
    """
    Launch the SciHub connector.
    """

    parser = argparse.ArgumentParser(formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    parser.add_argument("--config", default="", help="path to the config file")
    clargs = parser.parse_args()
    config_path = clargs.config

    config = ServiceConfig(config_path)

    sh = scihub.SciHub(config, RedisConnector)

    try:
        event = threading.Event()
        logger.success("Started SciHub connector")
        event.wait()
    except KeyboardInterrupt:
        sh.shutdown()


if __name__ == "__main__":
    main()



import abc
import time
from typing import Any

import numpy as np
import rich.progress
from rich.text import Text

from bec_lib import bec_logger

logger = bec_logger.logger


class MoveTaskProgressColumn(rich.progress.TaskProgressColumn):
    """Custom progress column for the move device progress bar"""

    def render(self, task) -> Text:
        """Renders the text for the progress bar"""
        if task.total is None and self.show_speed:
            return self.render_speed(task.finished_speed or task.speed)
        if task.fields.get("fields"):
            _text = f"[progress.percentage]{task.fields['fields'].get('current_pos'):10.2f} / {task.fields['fields'].get('target_pos'):10.2f} / {task.percentage:>3.0f} %"
        else:
            _text = f"[progress.percentage]{task.percentage:>3.0f}%"
        if self.markup:
            text = Text.from_markup(_text, style=self.style, justify=self.justify)
        else:
            text = Text(_text, style=self.style, justify=self.justify)
        if self.highlighter:
            self.highlighter.highlight(text)
        return text


class ScanTaskProgressColumn(rich.progress.TaskProgressColumn):
    """Custom progress column for the scan progress bar"""

    def render(self, task) -> Text:
        """Renders the text for the progress bar"""
        if task.total is None and self.show_speed:
            return self.render_speed(task.finished_speed or task.speed)
        if task.fields.get("fields"):
            _text = f"[progress.percentage]{int(task.fields['fields'].get('current_scan_pos'))} / {task.percentage:>3.0f} %"
        else:
            _text = f"[progress.percentage]{task.percentage:>3.0f}%"
        if self.markup:
            text = Text.from_markup(_text, style=self.style, justify=self.justify)
        else:
            text = Text(_text, style=self.style, justify=self.justify)
        if self.highlighter:
            self.highlighter.highlight(text)
        return text


class ProgressBarBase(abc.ABC):
    NUM_STEPS = 1000
    UPDATE_FREQUENCY = 10

    def __init__(self, clear_on_exit: bool = False) -> None:
        """Base class for progress bars. Override _init_tasks and _update_task for a new progress bar implementation.
        Override columns for a customized bar style.

        Args:
            clear_on_exit (bool, optional): remove progress bar after completion. Defaults to False.

        """
        self.clear_on_exit = clear_on_exit
        self._progress = None
        self._tasks = []

    @property
    def columns(self) -> tuple[rich.progress.ProgressColumn, ...]:
        """Columns used for a new Progress instance:
           - a text column for the description (TextColumn)
           - the bar itself (BarColumn)
           - a text column showing completion percentage (TextColumn)
           - an estimated-time-remaining column (TimeRemainingColumn)

        Override in subclasses to customize the progress bar appearance.

        Returns:
            tuple[rich.progress.ProgressColumn, ...]: columns
        """
        return rich.progress.Progress.get_default_columns()

    def start(self) -> None:
        """Start the Progress handler and initialize the tasks."""
        self._progress = rich.progress.Progress(
            *self.columns, transient=self.clear_on_exit, auto_refresh=False
        )
        self._progress.start()
        self._init_tasks()
        self._progress.refresh()

    def sleep(self):
        """Sleep for a short time to avoid busy waiting."""
        time.sleep(1 / self.UPDATE_FREQUENCY)

    def stop(self) -> None:
        """Stop the Progress handler"""
        self._progress.stop()

    @abc.abstractmethod
    def _init_tasks(self) -> None:
        """Initialize tasks by appending new items to self._progress."""

    @property
    def finished(self) -> bool:
        """True if all tasks have been completed.

        Returns:
            bool: True if all tasks have been completed.
        """
        return self._progress.finished

    @abc.abstractmethod
    def _update_task(self, task: int, value: Any) -> None:
        """Update routine that is applied to each tasks during self.update calls.

        Args:
            task (int): task ID
            value (Any): updated value received from self.update

        """

    def update(self, values: list):
        """Update the progress bar with new values.

        Args:
            values (list): list of values to update the progress bar with.
        """
        if not isinstance(values, list):
            values = [values]
        for i, task in enumerate(self._tasks):
            self._update_task(task=task, value=values[i])

    def __enter__(self):
        """Start the progress bar when entering a context."""
        self.start()
        return self

    def __exit__(self, *args):
        """Stop the progress bar when exiting a context."""
        self.stop()


class ScanProgressBar(ProgressBarBase):
    def __init__(self, scan_number: int, clear_on_exit=False) -> None:
        """Progress bar for a scan.

        Args:
            scan_number (int): scan number
            clear_on_exit (bool, optional): remove progress bar after completion. Defaults to False.
        """
        super().__init__(clear_on_exit)
        self._max_points = None
        self.scan_number = scan_number

    @property
    def columns(self) -> tuple:
        """Columns used for a new Progress instance.

        Returns:
            tuple: columns
        """
        return (
            rich.progress.TextColumn("[progress.description]{task.description}"),
            rich.progress.BarColumn(),
            ScanTaskProgressColumn(),
            rich.progress.TimeRemainingColumn(),
            rich.progress.TimeElapsedColumn(),
        )

    def _init_tasks(self):
        """Initialize tasks by appending new items to self._progress."""
        self._tasks.append(
            self._progress.add_task(f"[green] Scan {self.scan_number}: ", total=self.max_points)
        )

    def _update_tasks_total(self, max_points: int) -> None:
        """Update the total number of steps for the progress bar.

        Args:
            max_points (int): total number of steps
        """
        self._progress.tasks[0].total = max_points

    @property
    def max_points(self) -> int:
        """Total number of steps for the progress bar.

        Returns: int: total number of steps
        """
        return self._max_points

    @max_points.setter
    def max_points(self, max_points: int) -> None:
        """Setter for the total number of steps for the progress bar.

        Args:
            max_points (int): total number of steps
        """
        self._max_points = max_points
        self._update_tasks_total(max_points)

    def _update_task(self, task: int, value):
        """Update routine that is applied to each tasks during self.update calls.

        Args:
            task (int): task ID
            value (Any): updated value received from self.update
        """
        if self.max_points:
            self._progress.update(
                self._tasks[task], completed=value, fields={"current_scan_pos": value}, refresh=True
            )
            self._progress.refresh()
        else:
            self._progress.update(
                self._tasks[task], fields={"current_scan_pos": value}, refresh=True
            )
            self._progress.refresh()


class DeviceProgressBar(ProgressBarBase):
    def __init__(
        self,
        devices: list[str],
        target_values: list[float],
        start_values: list[float] = None,
        clear_on_exit: bool = False,
    ) -> None:
        """Progress bar for moving devices.

        Args:
            devices (list[str]): list of device names
            target_values (list[float]): list of target values
            start_values (list[float], optional): list of start values. Defaults to None.
            clear_on_exit (bool, optional): remove progress bar after completion. Defaults to False.
        """
        self.target_values = target_values
        self.start_values = start_values
        self.devices = devices

        super().__init__(clear_on_exit)

        self._tasks = []

    @property
    def columns(self):
        """Columns used for a new Progress instance."""
        return (
            rich.progress.TextColumn("[progress.description]{task.description}"),
            rich.progress.BarColumn(),
            MoveTaskProgressColumn(),
            rich.progress.TimeRemainingColumn(),
            rich.progress.TimeElapsedColumn(),
        )

    def _init_tasks(self):
        """Initialize tasks by appending new items to self._progress."""
        for ii, dev in enumerate(self.devices):
            self._tasks.append(
                self._progress.add_task(
                    f"[green] {dev}:{self.start_values[ii]:10.2f}", total=self.NUM_STEPS
                )
            )

    def _update_task(self, task: Any, value: float) -> None:
        """Update routine that is applied to each tasks during self.update calls.

        Args:
            task (Any): task ID
            value (float): updated value received from self.update
        """
        if self._progress.tasks[task].finished:
            return

        movement_range = self.target_values[task] - self.start_values[task]
        if np.abs(movement_range) > 0:
            completed = np.abs((value - self.start_values[task]) / movement_range * self.NUM_STEPS)
        else:
            completed = self.NUM_STEPS
        self._progress.update(
            task,
            completed=completed,
            fields={"current_pos": value, "target_pos": self.target_values[task]},
            refresh=True,
        )
        self._progress.refresh()

    def set_finished(self, device):
        """Set a progressbar for a device as finished.

        Args:
            device (str): device name
        """
        device_index = self.devices.index(device)
        self._progress.advance(self._tasks[device_index], self.NUM_STEPS)


if __name__ == "__main__":
    pass
    # devices = ["samx", "samy"]
    # target_values = [25, 50]
    # start = [0, 5]
    # steps_sim = []
    # for ii, dev in enumerate(devices):
    #     steps_sim.append(np.linspace(start[ii], target_values[ii], 100 * (ii + 1)))
    # loop_index = 0

    # def get_device_values(index):
    #     values = [
    #         steps_sim[dev_index][min(index, len(steps_sim[dev_index]) - 1)]
    #         for dev_index, _ in enumerate(devices)
    #     ]
    #     return values

    # with DeviceProgressBar(
    #     devices=devices, start_values=start, target_values=target_values
    # ) as progress:
    #     while not progress.finished:
    #         values = get_device_values(loop_index)
    #         progress.update(values=values)
    #         time.sleep(0.001)
    #         # for ii, dev in enumerate(devices):
    #         #     if np.isclose(values[ii], target_values[ii], atol=0.05):
    #         #         progress.set_finished(dev)
    #         loop_index += 1



from unittest import mock

from bec_ipython_client.callbacks.scan_progress import LiveUpdatesScanProgress
from bec_lib import messages


def test_update_progressbar_continues_without_device_data():
    bec = mock.MagicMock()
    request = mock.MagicMock()
    live_update = LiveUpdatesScanProgress(bec=bec, report_instruction={}, request=request)
    progressbar = mock.MagicMock()

    bec.connector.get.return_value = None
    res = live_update._update_progressbar(progressbar, "async_dev1")
    assert res is False


def test_update_progressbar_continues_when_scan_id_doesnt_match():
    bec = mock.MagicMock()
    request = mock.MagicMock()
    live_update = LiveUpdatesScanProgress(bec=bec, report_instruction={}, request=request)
    progressbar = mock.MagicMock()
    live_update.scan_item = mock.MagicMock()
    live_update.scan_item.scan_id = "scan_id2"

    bec.connector.get.return_value = messages.ProgressMessage(
        value=1, max_value=10, done=False, metadata={"scan_id": "scan_id"}
    )
    res = live_update._update_progressbar(progressbar, "async_dev1")
    assert res is False


def test_update_progressbar_updates_max_value():
    bec = mock.MagicMock()
    request = mock.MagicMock()
    live_update = LiveUpdatesScanProgress(bec=bec, report_instruction={}, request=request)
    progressbar = mock.MagicMock()
    live_update.scan_item = mock.MagicMock()
    live_update.scan_item.scan_id = "scan_id"

    bec.connector.get.return_value = messages.ProgressMessage(
        value=10, max_value=20, done=False, metadata={"scan_id": "scan_id"}
    )
    res = live_update._update_progressbar(progressbar, "async_dev1")
    assert res is False
    assert progressbar.max_points == 20
    progressbar.update.assert_called_once_with(10)


def test_update_progressbar_returns_true_when_max_value_is_reached():
    bec = mock.MagicMock()
    request = mock.MagicMock()
    live_update = LiveUpdatesScanProgress(bec=bec, report_instruction={}, request=request)
    progressbar = mock.MagicMock()
    live_update.scan_item = mock.MagicMock()
    live_update.scan_item.scan_id = "scan_id"

    bec.connector.get.return_value = messages.ProgressMessage(
        value=10, max_value=10, done=True, metadata={"scan_id": "scan_id"}
    )
    res = live_update._update_progressbar(progressbar, "async_dev1")
    assert res is True



import time

from bec_lib import bec_logger
from bec_server.scan_server.scans import ScanArgType, ScanBase, SyncFlyScanBase

logger = bec_logger.logger


class OTFScan(SyncFlyScanBase):
    scan_name = "otf_scan"
    scan_report_hint = "table"
    required_kwargs = ["e1", "e2", "time"]
    arg_input = {}
    arg_bundle_size = {"bundle": len(arg_input), "min": None, "max": None}

    def __init__(self, *args, parameter: dict = None, **kwargs):
        """Scans the energy from e1 to e2 in <time> minutes.

        Examples:
            >>> scans.otf_scan(e1=700, e2=740, time=4)

        """
        super().__init__(parameter=parameter, **kwargs)
        self.axis = []
        self.scan_motors = []
        self.num_pos = 0
        self.mono = self.caller_kwargs.get("mono", "mono")
        self.otf_device = self.caller_kwargs.get("otf", "otf")

    def pre_scan(self):
        yield None

    @property
    def monitor_sync(self):
        return self.otf_device

    def scan_core(self):
        yield from self.stubs.set(
            device=self.mono, value=self.caller_kwargs["e1"], wait_group="flyer"
        )
        yield from self.stubs.wait(device=[self.mono], wait_group="flyer", wait_type="move")
        yield from self.stubs.kickoff(
            device=self.otf_device,
            parameter={
                key: val for key, val in self.caller_kwargs.items() if key in ["e1", "e2", "time"]
            },
        )
        yield from self.stubs.wait(device=[self.otf_device], wait_group="kickoff", wait_type="move")
        yield from self.stubs.complete(device=self.otf_device)
        target_diid = self.DIID - 1

        while True:
            yield from self.stubs.read_and_wait(group="primary", wait_group="readout_primary")
            status = self.stubs.get_req_status(
                device=self.otf_device, RID=self.metadata["RID"], DIID=target_diid
            )
            progress = self.stubs.get_device_progress(
                device=self.otf_device, RID=self.metadata["RID"]
            )
            if progress:
                self.num_pos = progress
            if status:
                break
            time.sleep(1)


class HystScan(ScanBase):
    scan_name = "hyst_scan"
    scan_report_hint = "table"
    required_kwargs = []
    arg_input = {
        "field_motor": ScanArgType.DEVICE,
        "start_field": ScanArgType.FLOAT,
        "end_field": ScanArgType.FLOAT,
        "mono": ScanArgType.DEVICE,
        "energy1": ScanArgType.FLOAT,
        "energy2": ScanArgType.FLOAT,
    }
    arg_bundle_size = {"bundle": 3, "min": 1, "max": 1}
    scan_type = "step"
    default_ramp_rate = 2

    def __init__(self, *args, parameter: dict = None, **kwargs):
        """
        A hysteresis scan.

        scans.hyst_scan(field_motor, start_field, end_field, mono, energy1, energy2)

        Examples:
            >>> scans.hyst_scan(dev.field_x, 0, 0.5, dev.mono, 600, 640, ramp_rate=2)

        """
        super().__init__(parameter=parameter, **kwargs)
        self.axis = []
        self.flyer = list(self.caller_args.keys())[0]
        self.energy_motor = list(self.caller_args.keys())[1]
        self.scan_motors = [self.energy_motor, self.flyer]
        self.flyer_positions = self.caller_args[self.flyer]
        self._current_scan_motor_index = 0
        self._scan_motor_direction = 1
        self.ramp_rate = self.caller_kwargs.get("ramp_rate", self.default_ramp_rate)

    def _calculate_positions(self) -> None:
        self.positions = [[pos] for pos in self.caller_args[self.energy_motor]]

    def prepare_positions(self):
        self._calculate_positions()
        self.num_pos = 0
        yield None
        self._check_limits()

    def _at_each_point(self):
        yield from self.stubs.read(group="primary", wait_group="primary", point_id=self.point_id)
        self.point_id += 1

    def _get_next_scan_motor_position(self):
        while True:
            yield self.positions[self._current_scan_motor_index][0]

            if len(self.positions) - 1 == self._current_scan_motor_index or (
                self._current_scan_motor_index == 0 and self._scan_motor_direction < 0
            ):
                self._scan_motor_direction *= -1
            self._current_scan_motor_index += self._scan_motor_direction

    def scan_core(self):
        # yield from self._move_and_wait(self.positions[0])
        status = yield from self.stubs.send_rpc_and_wait(
            "field_x", "ramprate.set", self.default_ramp_rate
        )
        status.wait()
        yield from self.stubs.set(
            device=self.flyer, value=self.flyer_positions[0], wait_group="flyer"
        )
        yield from self.stubs.wait(device=[self.flyer], wait_group="flyer", wait_type="move")

        status = yield from self.stubs.send_rpc_and_wait("field_x", "ramprate.set", self.ramp_rate)
        status.wait()
        # send the slow motor on its way
        yield from self.stubs.set(
            device=self.flyer, value=self.flyer_positions[1], wait_group="flyer"
        )

        flyer_done = False
        pos_generator = self._get_next_scan_motor_position()
        target_DIID = self.DIID - 1
        dev = self.device_manager.devices
        while not flyer_done:
            flyer_done = bool(
                self.stubs.get_req_status(
                    device=self.flyer, DIID=target_DIID, RID=self.metadata["RID"]
                )
            )

            val = next(pos_generator)
            logger.info(f"Moving mono to {val}.")
            yield from self.stubs.set(device=self.energy_motor, value=val, wait_group="mono")
            yield from self.stubs.wait(
                device=[self.energy_motor], wait_group="mono", wait_type="move"
            )

            monitored_devices = [dev.name for dev in dev.monitored_devices([])]
            yield from self.stubs.read_and_wait(
                device=[self.flyer, self.scan_motors[0], *monitored_devices],
                wait_group="readout_primary",
                point_id=self.point_id,
            )
            # time.sleep(1)
            self.point_id += 1
            self.scan_motors[0]
            self.num_pos += 1

        status = yield from self.stubs.send_rpc_and_wait(
            "field_x", "ramprate.set", self.default_ramp_rate
        )
        status.wait()

    def return_to_start(self):
        yield None



from unittest import mock

import numpy as np
import pytest
from py_scibec_openapi_client.models import NewScanData, ScanPartial

from bec_lib import messages
from bec_lib.redis_connector import MessageObject
from bec_server.scihub.scibec.scibec_metadata_handler import SciBecMetadataHandler


@pytest.fixture
def md_handler():
    inst = SciBecMetadataHandler(mock.Mock())
    yield inst


def test_handle_scan_status(md_handler):
    # pylint: disable=protected-access
    msg = messages.ScanStatusMessage(scan_id="scan_id", status="open", info={})
    with mock.patch.object(md_handler, "update_scan_status") as mock_update_scan_status:
        md_handler._handle_scan_status(
            MessageObject(value=msg, topic="scan_status"), parent=md_handler
        )
        mock_update_scan_status.assert_called_once_with(msg)


def test_handle_scan_status_ignores_errors(md_handler):
    # pylint: disable=protected-access
    msg = messages.ScanStatusMessage(scan_id="scan_id", status="open", info={})
    with mock.patch("bec_server.scihub.scibec.scibec_metadata_handler.logger") as mock_logger:
        with mock.patch.object(md_handler, "update_scan_status") as mock_update_scan_status:
            mock_update_scan_status.side_effect = Exception("test")
            md_handler._handle_scan_status(
                MessageObject(value=msg, topic="scan_status"), parent=md_handler
            )
            mock_update_scan_status.assert_called_once_with(msg)
            mock_logger.exception.assert_called_once_with(
                f"Failed to update scan status: {Exception('test')}"
            )


def test_update_scan_status_returns_without_scibec(md_handler):
    # pylint: disable=protected-access
    msg = messages.ScanStatusMessage(scan_id="scan_id", status="open", info={})
    md_handler.scibec_connector.scibec = None
    md_handler.update_scan_status(msg)


def test_update_scan_status(md_handler, active_experiment, dataset_document):
    # pylint: disable=protected-access
    msg = messages.ScanStatusMessage(scan_id="scan_id", status="open", info={"dataset_number": 12})
    scibec = mock.Mock()
    md_handler.scibec_connector.scibec = scibec
    scibec_info = {"activeExperiment": active_experiment}
    md_handler.scibec_connector.scibec_info = scibec_info
    scibec.scan.scan_controller_find = mock.MagicMock(return_value=[])
    scibec.dataset.dataset_controller_find = mock.MagicMock(return_value=[])
    scibec.dataset.dataset_controller_create = mock.MagicMock(return_value=dataset_document)
    md_handler.update_scan_status(msg)


def test_update_scan_status_patch(md_handler, active_experiment, scan_document):
    # pylint: disable=protected-access
    msg = messages.ScanStatusMessage(
        scan_id="scan_id", status="closed", info={"dataset_number": 12}
    )
    scibec = mock.Mock()
    md_handler.scibec_connector.scibec = scibec
    scibec_info = {"activeExperiment": active_experiment}
    md_handler.scibec_connector.scibec_info = scibec_info
    scibec.scan.scan_controller_find = mock.MagicMock(return_value=[scan_document])
    md_handler.update_scan_status(msg)
    scibec.scan.scan_controller_update_by_id.assert_called_once_with(
        id="dummy_id",
        scan_partial=ScanPartial(exit_status="closed", metadata={"dataset_number": 12}),
    )


def test_handle_file_content(md_handler):
    # pylint: disable=protected-access
    msg = messages.FileContentMessage(file_path="my_file.h5", data={"data": {}})
    msg_raw = MessageObject(value=msg, topic="file_content")
    with mock.patch.object(md_handler, "update_scan_data") as mock_update_scan_data:
        md_handler._handle_file_content(msg_raw, parent=md_handler)
        mock_update_scan_data.assert_called_once_with(**msg.content)


def test_handle_file_content_ignores_errors(md_handler):
    # pylint: disable=protected-access
    msg = messages.FileContentMessage(file_path="my_file.h5", data={"data": {}})
    msg_raw = MessageObject(value=msg, topic="file_content")
    with mock.patch("bec_server.scihub.scibec.scibec_metadata_handler.logger") as mock_logger:
        with mock.patch.object(md_handler, "update_scan_data") as mock_update_scan_data:
            mock_update_scan_data.side_effect = Exception("test")
            md_handler._handle_file_content(msg_raw, parent=md_handler)
            mock_update_scan_data.assert_called_once_with(**msg.content)
            mock_logger.exception.assert_called_once_with(
                f"Failed to update scan data: {Exception('test')}"
            )


def test_update_scan_data_return_without_scibec(md_handler):
    # pylint: disable=protected-access
    md_handler.scibec_connector.scibec = None
    md_handler.update_scan_data(file_path="my_file.h5", data={"data": {}})


def test_update_scan_data_without_scan(md_handler):
    # pylint: disable=protected-access
    scibec = mock.Mock()
    md_handler.scibec_connector.scibec = scibec
    scibec.scan.scan_controller_find = mock.MagicMock(return_value=[])
    md_handler.update_scan_data(
        file_path="my_file.h5", data={"data": {}, "metadata": {"scan_id": "scan_id"}}
    )


def test_update_scan_data(md_handler, scan_document):
    # pylint: disable=protected-access
    scibec = mock.Mock()
    md_handler.scibec_connector.scibec = scibec
    scibec.scan.scan_controller_find = mock.MagicMock(return_value=[scan_document])
    md_handler.update_scan_data(
        file_path="my_file.h5", data={"data": {}, "metadata": {"scan_id": "scan_id"}}
    )
    scibec.scan_data.scan_data_controller_create_many.assert_called_once_with(
        NewScanData(
            **{
                "readACL": ["readACL"],
                "writeACL": ["readACL"],
                "owner": ["owner"],
                "scanId": "dummy_id",
                "filePath": "my_file.h5",
                "data": {"data": {}, "metadata": {"scan_id": "scan_id"}},
            }
        )
    )


def test_update_scan_data_exceeding_limit(md_handler, scan_document):
    # pylint: disable=protected-access
    scibec = mock.Mock()
    md_handler.MAX_DATA_SIZE = 1000
    md_handler.scibec_connector.scibec = scibec
    scibec.scan.scan_controller_find = mock.MagicMock(return_value=[scan_document])
    data_block = {f"key_{i}": {"signal": list(range(100))} for i in range(10)}
    data_block.update({"metadata": {"scan_id": "scan_id"}})
    md_handler.update_scan_data(file_path="my_file.h5", data=data_block)
    num_calls = scibec.scan_data.scan_data_controller_create_many.call_count
    assert num_calls == 5


@pytest.mark.parametrize(
    "data, expected_result",
    [
        (
            {"int": 123, "float": 3.14, "str": "hello", "bool": True},
            {"int": 123, "float": 3.14, "str": "hello", "bool": True},
        ),
        (
            {"nested": {"int": 123, "float": 3.14, "str": "hello", "bool": True}},
            {"nested": {"int": 123, "float": 3.14, "str": "hello", "bool": True}},
        ),
        ([1, 2, 3, 4, 5], [1, 2, 3, 4, 5]),
        ((1, 2, 3, 4, 5), (1, 2, 3, 4, 5)),
        ({1, 2, 3, 4, 5}, {1, 2, 3, 4, 5}),
        (
            np.array([1, 2, 3, 4, 5]),
            '{"nd": true, "type": "<i8", "kind": "", "shape": [5], "data": [1, 2, 3, 4, 5]}',
        ),
        # Integer types
        ({"signal": np.int8(123)}, {"signal": 123}),
        ({"signal": np.int16(123)}, {"signal": 123}),
        ({"signal": np.int32(123)}, {"signal": 123}),
        ({"signal": np.int64(123)}, {"signal": 123}),
        ({"signal": np.uint8(123)}, {"signal": 123}),
        ({"signal": np.uint16(123)}, {"signal": 123}),
        ({"signal": np.uint32(123)}, {"signal": 123}),
        ({"signal": np.uint64(123)}, {"signal": 123}),
        # Float types
        ({"signal": np.float16(3.14)}, {"signal": np.float16(3.14).tolist()}),
        ({"signal": np.float32(3.14)}, {"signal": np.float32(3.14).tolist()}),
        ({"signal": np.float64(3.14)}, {"signal": np.float64(3.14).tolist()}),
        ({"signal": np.str_("hello")}, {"signal": "hello"}),
        ({"signal": np.bool_(True)}, {"signal": True}),
    ],
)
def test_serialize_special_data(md_handler, data, expected_result):
    # pylint: disable=protected-access
    assert md_handler.serialize_special_data(data) == expected_result



import numpy as np
import pydantic
import pytest

from bec_lib import messages
from bec_lib.serialization import MsgpackSerialization


@pytest.mark.parametrize("version", [1.0, 1.1, 1.2, None])
def test_bec_message_msgpack_serialization_version(version):
    msg = messages.DeviceInstructionMessage(
        device="samx", action="set", parameter={"set": 0.5}, metadata={"RID": "1234"}
    )
    if version is not None and version < 1.2:
        with pytest.raises(RuntimeError) as exception:
            MsgpackSerialization.dumps(msg, version=version)
        assert "Unsupported BECMessage version" in str(exception.value)
    else:
        res = MsgpackSerialization.dumps(msg)
        print(res)
        v12res = b'\xc7z\x01BECMSG_1.2_34_67_EOH_{"msg_type": "device_instruction"}\x84\xa8metadata\x81\xa3RID\xa41234\xa6device\xa4samx\xa6action\xa3set\xa9parameter\x81\xa3set\xcb?\xe0\x00\x00\x00\x00\x00\x00'
        assert res == v12res
        res_loaded = MsgpackSerialization.loads(res)
        assert res_loaded == msg


@pytest.mark.parametrize("version", [1.2, None])
def test_bec_message_serialization_numpy_ndarray(version):
    msg = messages.DeviceMessage(
        signals={"samx": {"value": np.random.rand(20).astype(np.float32)}}, metadata={"RID": "1234"}
    )
    res = MsgpackSerialization.dumps(msg)
    print(res)
    res_loaded = MsgpackSerialization.loads(res)
    np.testing.assert_equal(res_loaded.content, msg.content)
    assert res_loaded == msg


def test_bundled_message():
    sub_msg = messages.DeviceMessage(signals={"samx": {"value": 5.2}}, metadata={"RID": "1234"})
    msg = messages.BundleMessage()
    msg.append(sub_msg)
    msg.append(sub_msg)
    res = MsgpackSerialization.dumps(msg)
    res_loaded = MsgpackSerialization.loads(res)
    assert res_loaded == [sub_msg, sub_msg]


def test_ScanQueueModificationMessage():
    msg = messages.ScanQueueModificationMessage(
        scan_id="1234", action="halt", parameter={"RID": "1234"}
    )
    res = MsgpackSerialization.dumps(msg)
    res_loaded = MsgpackSerialization.loads(res)
    assert res_loaded == msg


def test_ScanQueueModificationMessage_with_wrong_action_returns_None():
    with pytest.raises(pydantic.ValidationError):
        messages.ScanQueueModificationMessage(
            scan_id="1234", action="wrong_action", parameter={"RID": "1234"}
        )


def test_ScanQueueStatusMessage_must_include_primary_queue():
    with pytest.raises(pydantic.ValidationError):
        messages.ScanQueueStatusMessage(queue={}, metadata={"RID": "1234"})


def test_ScanQueueStatusMessage_loads_successfully():
    msg = messages.ScanQueueStatusMessage(queue={"primary": {}}, metadata={"RID": "1234"})
    res = MsgpackSerialization.dumps(msg)
    res_loaded = MsgpackSerialization.loads(res)
    assert res_loaded == msg


def test_DeviceMessage_loads_successfully():
    msg = messages.DeviceMessage(signals={"samx": {"value": 5.2}}, metadata={"RID": "1234"})
    res = MsgpackSerialization.dumps(msg)
    res_loaded = MsgpackSerialization.loads(res)
    assert res_loaded == msg


def test_DeviceMessage_must_include_signals_as_dict():
    with pytest.raises(pydantic.ValidationError):
        messages.DeviceMessage(signals="wrong_signals", metadata={"RID": "1234"})


def test_ClientInfoMessage():
    msg = messages.ClientInfoMessage(
        message="test", show_asap=True, RID="1234", metadata={"RID": "1234"}
    )
    res = MsgpackSerialization.dumps(msg)
    res_loaded = MsgpackSerialization.loads(res)
    assert res_loaded == msg


def test_ClientInfoMessage_raises():
    with pytest.raises(pydantic.ValidationError):
        messages.ClientInfoMessage(
            message="test",
            source="abc",
            show_asap=True,
            RID="1234",
            metadata={"RID": "1234", "wrong": "wrong"},
        )


def test_DeviceRPCMessage():
    msg = messages.DeviceRPCMessage(
        device="samx", return_val=1, out="done", success=True, metadata={"RID": "1234"}
    )
    res = MsgpackSerialization.dumps(msg)
    res_loaded = MsgpackSerialization.loads(res)
    assert res_loaded == msg


def test_DeviceStatusMessage():
    msg = messages.DeviceStatusMessage(device="samx", status=0, metadata={"RID": "1234"})
    res = MsgpackSerialization.dumps(msg)
    res_loaded = MsgpackSerialization.loads(res)
    assert res_loaded == msg


def test_DeviceReqStatusMessage():
    msg = messages.DeviceReqStatusMessage(device="samx", success=True, metadata={"RID": "1234"})
    res = MsgpackSerialization.dumps(msg)
    res_loaded = MsgpackSerialization.loads(res)
    assert res_loaded == msg


def test_DeviceInfoMessage():
    msg = messages.DeviceInfoMessage(
        device="samx", info={"version": "1.0"}, metadata={"RID": "1234"}
    )
    res = MsgpackSerialization.dumps(msg)
    res_loaded = MsgpackSerialization.loads(res)
    assert res_loaded == msg


def test_ScanMessage():
    msg = messages.ScanMessage(
        point_id=1, scan_id="scan_id", data={"value": 3}, metadata={"RID": "1234"}
    )
    res = MsgpackSerialization.dumps(msg)
    res_loaded = MsgpackSerialization.loads(res)
    assert res_loaded == msg


def test_ScanBaselineMessage():
    msg = messages.ScanBaselineMessage(
        scan_id="scan_id", data={"value": 3}, metadata={"RID": "1234"}
    )
    res = MsgpackSerialization.dumps(msg)
    res_loaded = MsgpackSerialization.loads(res)
    assert res_loaded == msg


@pytest.mark.parametrize(
    "action,valid",
    [("add", True), ("set", True), ("update", True), ("reload", True), ("wrong_action", False)],
)
def test_DeviceConfigMessage(action, valid):
    if valid:
        msg = messages.DeviceConfigMessage(
            action=action, config={"device": "samx"}, metadata={"RID": "1234"}
        )
        res = MsgpackSerialization.dumps(msg)
        res_loaded = MsgpackSerialization.loads(res)
        assert res_loaded == msg
    else:
        with pytest.raises(pydantic.ValidationError):
            messages.DeviceConfigMessage(
                action=action, config={"device": "samx"}, metadata={"RID": "1234"}
            )


def test_LogMessage():
    msg = messages.LogMessage(
        log_type="error", log_msg="An error occurred", metadata={"RID": "1234"}
    )
    res = MsgpackSerialization.dumps(msg)
    res_loaded = MsgpackSerialization.loads(res)
    assert res_loaded == msg


def test_AlarmMessage():
    msg = messages.AlarmMessage(
        severity=2,
        alarm_type="major",
        source={"system": "samx"},
        msg="An error occurred",
        metadata={"RID": "1234"},
    )
    res = MsgpackSerialization.dumps(msg)
    res_loaded = MsgpackSerialization.loads(res)
    assert res_loaded == msg


def test_StatusMessage():
    msg = messages.StatusMessage(
        name="system",
        status=messages.BECStatus.RUNNING,
        info={"version": "1.0"},
        metadata={"RID": "1234"},
    )
    res = MsgpackSerialization.dumps(msg)
    res_loaded = MsgpackSerialization.loads(res)
    assert res_loaded == msg


def test_FileMessage():
    msg = messages.FileMessage(
        file_path="/path/to/file", done=True, successful=True, metadata={"RID": "1234"}
    )
    res = MsgpackSerialization.dumps(msg)
    res_loaded = MsgpackSerialization.loads(res)
    assert res_loaded == msg


def test_VariableMessage():
    msg = messages.VariableMessage(value="value", metadata={"RID": "1234"})
    res = MsgpackSerialization.dumps(msg)
    res_loaded = MsgpackSerialization.loads(res)
    assert res_loaded == msg


def test_ObserverMessage():
    msg = messages.ObserverMessage(observer=[{"name": "observer1"}], metadata={"RID": "1234"})
    res = MsgpackSerialization.dumps(msg)
    res_loaded = MsgpackSerialization.loads(res)
    assert res_loaded == msg


def test_ServiceMetricMessage():
    msg = messages.ServiceMetricMessage(
        name="service1", metrics={"metric1": 1}, metadata={"RID": "1234"}
    )
    res = MsgpackSerialization.dumps(msg)
    res_loaded = MsgpackSerialization.loads(res)
    assert res_loaded == msg


def test_ProcessedDataMessage():
    msg = messages.ProcessedDataMessage(data={"samx": {"value": 5.2}}, metadata={"RID": "1234"})
    res = MsgpackSerialization.dumps(msg)
    res_loaded = MsgpackSerialization.loads(res)
    assert res_loaded == msg


def test_DAPConfigMessage():
    msg = messages.DAPConfigMessage(config={"val": "val"}, metadata={"RID": "1234"})
    res = MsgpackSerialization.dumps(msg)
    res_loaded = MsgpackSerialization.loads(res)
    assert res_loaded == msg


def test_AvailableResourceMessage():
    msg = messages.AvailableResourceMessage(
        resource={"resource": "available"}, metadata={"RID": "1234"}
    )
    res = MsgpackSerialization.dumps(msg)
    res_loaded = MsgpackSerialization.loads(res)
    assert res_loaded == msg


def test_ProgressMessage():
    msg = messages.ProgressMessage(value=0.5, max_value=10, done=False, metadata={"RID": "1234"})
    res = MsgpackSerialization.dumps(msg)
    res_loaded = MsgpackSerialization.loads(res)
    assert res_loaded == msg


def test_GUIConfigMessage():
    msg = messages.GUIConfigMessage(config={"config": "value"}, metadata={"RID": "1234"})
    res = MsgpackSerialization.dumps(msg)
    res_loaded = MsgpackSerialization.loads(res)
    assert res_loaded == msg


def test_ScanQueueHistoryMessage():
    msg = messages.ScanQueueHistoryMessage(
        status="running", queue_id="queue_id", info={"val": "val"}, metadata={"RID": "1234"}
    )
    res = MsgpackSerialization.dumps(msg)
    res_loaded = MsgpackSerialization.loads(res)
    assert res_loaded == msg


def test_DAPResponseMessage():
    msg = messages.DAPResponseMessage(success=True, data=({}, None), metadata={"RID": "1234"})
    res = MsgpackSerialization.dumps(msg)
    res_loaded = MsgpackSerialization.loads(res)
    assert res_loaded == msg


def test_DAPResponseMessage_accepts_None():
    msg = messages.DAPResponseMessage(success=True, data=None, metadata={"RID": "1234"})
    res = MsgpackSerialization.dumps(msg)
    res_loaded = MsgpackSerialization.loads(res)
    assert res_loaded == msg


def test_DAPRequestMessage():
    msg = messages.DAPRequestMessage(
        dap_cls="dap_cls",
        dap_type="continuous",
        config={"config": "value"},
        metadata={"RID": "1234"},
    )
    res = MsgpackSerialization.dumps(msg)
    res_loaded = MsgpackSerialization.loads(res)
    assert res_loaded == msg


def test_wrong_DAPRequestMessage():
    with pytest.raises(pydantic.ValidationError):
        messages.DAPRequestMessage(
            dap_cls="dap_cls",
            dap_type="error",
            config={"config": "value"},
            metadata={"RID": "1234"},
        )


def test_FileContentMessage():
    msg = messages.FileContentMessage(file_path="/path/to/file", data={}, metadata={"RID": "1234"})
    res = MsgpackSerialization.dumps(msg)
    res_loaded = MsgpackSerialization.loads(res)
    assert res_loaded == msg


def test_CredentialsMessage():
    msg = messages.CredentialsMessage(credentials={"username": "user", "password": "pass"})
    res = MsgpackSerialization.dumps(msg)
    res_loaded = MsgpackSerialization.loads(res)
    assert res_loaded == msg


def test_DeviceInstructionMessage():
    msg = messages.DeviceInstructionMessage(
        device="samx", action="set", parameter={"set": 0.5}, metadata=None
    )
    res = MsgpackSerialization.dumps(msg)
    res_loaded = MsgpackSerialization.loads(res)
    assert res_loaded == msg
    assert res_loaded.metadata == {}



# Description: Launch the scan bundler.
# This script is the entry point for the Scan Bundler. It is called either
# by the bec-scan-bundler entry point or directly from the command line.
import argparse
import threading

from bec_lib import RedisConnector, ServiceConfig, bec_logger
from bec_server import scan_bundler

logger = bec_logger.logger
bec_logger.level = bec_logger.LOGLEVEL.INFO


def main():
    """
    Launch the scan bundler.
    """
    parser = argparse.ArgumentParser(formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    parser.add_argument("--config", default="", help="path to the config file")
    clargs = parser.parse_args()
    config_path = clargs.config

    config = ServiceConfig(config_path)

    sb = scan_bundler.ScanBundler(config, RedisConnector)

    try:
        event = threading.Event()
        logger.success("Started ScanBundler")
        event.wait()
    except KeyboardInterrupt:
        sb.shutdown()


if __name__ == "__main__":
    main()



import importlib
import inspect
from functools import lru_cache
from typing import Literal

from bec_lib.logger import bec_logger

logger = bec_logger.logger


def get_plugin_class(class_spec: str, additional_modules=None) -> type:
    """
    Load a class from a plugin module.
    The class specification should follow the syntax <package>.<module>.<class> and is thus
    equivalent to the import statement `from <package>.<module> import <class>`. Imports along
    the lines of `from <package> import <class>` are also supported, assuming that the class is
    residing in the `__init__.py` file of the package.

    Args:
        class_spec (str): The class specification.
        additional_modules (list): Additional modules to search for the class.

    Returns:
        class: The class object.

    Raises:
        ValueError: If the class specification is invalid.
        ModuleNotFoundError: If the module could not be found.
    """
    class_specs = class_spec.split(".")
    if len(class_specs) == 1:
        raise ValueError(
            f"Invalid specification {class_spec}: The class spec should follow the syntax <package>.<module>.<class>"
        )

    parent_module = class_specs[0]
    full_module = ".".join(class_specs[:-1])
    class_name = class_specs[-1]

    if parent_module not in [
        plugins.__name__ for plugins in _get_available_modules("bec", additional_modules)
    ]:
        raise ModuleNotFoundError(f"Could not find module {parent_module}.")

    module = _import_module(full_module)
    return getattr(module, class_name)


def get_scan_plugins() -> dict:
    """
    Load all scan plugins.

    Returns:
        dict: A dictionary with the plugin names as keys and the plugin classes as values.
    """
    modules = _get_available_plugins("bec.scans")
    loaded_plugins = {}
    for module in modules:
        mods = inspect.getmembers(module, predicate=_filter_plugins)
        for name, mod_cls in mods:
            if name in loaded_plugins:
                logger.warning(f"Duplicated scan plugin {name}.")
            loaded_plugins[name] = mod_cls
    return loaded_plugins


def get_file_writer_plugins() -> dict:
    """
    Load all file writer plugins.

    Returns:
        dict: A dictionary with the plugin names as keys and the plugin classes as values.
    """
    modules = _get_available_plugins("bec.file_writer")
    loaded_plugins = {}
    for module in modules:
        mods = inspect.getmembers(module, predicate=_filter_plugins)
        for name, mod_cls in mods:
            if name in loaded_plugins:
                logger.warning(f"Duplicated file writer plugin {name}.")
            loaded_plugins[name] = mod_cls
    return loaded_plugins


def get_ipython_client_startup_plugins(state: Literal["pre", "post"]) -> dict:
    """
    Load all IPython client startup plugins.

    Args:
        state (str): The state of the plugin. Either "pre" or "post".

    Returns:
        dict: A dictionary with the plugin names as keys and the plugin module and source name as values.
    """
    group = "bec.ipython_client_startup"
    target = f"plugin_ipython_client_{state}"
    entry_points = importlib.metadata.entry_points(group=group)
    modules = {
        entry_point.name: {"source": entry_point.dist.name, "module": entry_point.load()}
        for entry_point in entry_points
        if entry_point.name == target
    }
    return modules


def _filter_plugins(module) -> bool:
    """
    Filter out classes that are not plugins.

    Args:
        module: The module to filter.

    Returns:
        bool: True if the module is a plugin, False otherwise.
    """
    return inspect.isclass(module) and not module.__name__.startswith("__")


@lru_cache(maxsize=10)
def _get_available_plugins(group) -> list:
    """
    Retrieve all available plugins for a given plugin group.

    Args:
        group (str): The name of the group.

    Returns:
        list: A list of modules.
    """

    plugins = importlib.metadata.entry_points(group=group)
    modules = [plugin.load() for plugin in plugins]
    return modules


def _get_available_modules(plugin: str, additional_modules=None) -> list:
    """
    Load all available modules for a given plugin.

    Args:
        plugin (str): The name of the plugin.
        additional_modules (list): Additional modules to append to the list of available modules.

    Returns:
        list: A list of modules.
    """

    modules = _get_available_plugins(plugin)
    if additional_modules:
        modules.extend(additional_modules)
    return modules


@lru_cache(maxsize=20)
def _import_module(module_name: str):
    """
    Import a module by name.
    """
    module = importlib.import_module(module_name)
    return module



import argparse
import os

import libtmux

from bec_server.bec_server_utils.service_handler import ServiceHandler


def main():
    """
    Launch the BEC server in a tmux session. All services are launched in separate panes.
    """
    parser = argparse.ArgumentParser(description="Utility tool managing the BEC server")
    command = parser.add_subparsers(dest="command")
    start = command.add_parser("start", help="Start the BEC server")
    start.add_argument(
        "--config", type=str, default=None, help="Path to the BEC service config file"
    )
    start.add_argument(
        "--no-tmux", action="store_true", default=False, help="Do not start processes in tmux"
    )
    command.add_parser("stop", help="Stop the BEC server")
    restart = command.add_parser("restart", help="Restart the BEC server")
    restart.add_argument(
        "--config", type=str, default=None, help="Path to the BEC service config file"
    )
    command.add_parser("attach", help="Open the currently running BEC server session")

    args = parser.parse_args()
    try:
        # 'stop' has no config
        config = args.config
    except AttributeError:
        config = None

    service_handler = ServiceHandler(
        bec_path=os.path.dirname(os.path.dirname(os.path.dirname(os.path.abspath(__file__)))),
        config_path=config,
        no_tmux=args.no_tmux if "no_tmux" in args else False,
    )
    if args.command == "start":
        service_handler.start()
    elif args.command == "stop":
        service_handler.stop()
    elif args.command == "restart":
        service_handler.restart()
    elif args.command == "attach":
        server = libtmux.Server()
        session = server.find_where({"session_name": "bec"})
        if session is None:
            print("No BEC session found")
            return
        session.attach_session()



from .ophyd_patch import monkey_patch_ophyd

monkey_patch_ophyd()

from .devices.sls_devices import SLSInfo, SLSOperatorMessages
from .sim.sim import SimCamera
from .sim.sim import SimFlyer
from .sim.sim import SimFlyer as SynFlyer
from .sim.sim import SimMonitor
from .sim.sim import SimMonitor as SynAxisMonitor
from .sim.sim import SimMonitor as SynGaussBEC
from .sim.sim import SimPositioner
from .sim.sim import SimPositioner as SynAxisOPAAS
from .sim.sim import SimWaveform, SynDeviceOPAAS
from .sim.sim_frameworks import DeviceProxy, H5ImageReplayProxy, SlitProxy
from .sim.sim_signals import ReadOnlySignal
from .sim.sim_signals import ReadOnlySignal as SynSignalRO
from .utils.bec_device_base import BECDeviceBase
from .utils.dynamic_pseudo import ComputedSignal
from .utils.static_device_test import launch



"""
Threading utility functions for the bec_lib package.
"""

import functools


def threadlocked(fcn):
    """Ensure that the thread acquires and releases the lock."""

    @functools.wraps(fcn)
    def wrapper(self, *args, **kwargs):
        # pylint: disable=protected-access
        with self._lock:
            return fcn(self, *args, **kwargs)

    return wrapper



"""
This module provides a command line interface to monitor a channel.
"""

import argparse
import json
import re
import threading

from bec_lib.endpoints import MessageEndpoints
from bec_lib.redis_connector import RedisConnector


def channel_callback(msg, **_kwargs):
    """
    Callback for channel monitor.
    """
    msg = msg.value
    out = {"msg_type": msg.msg_type, "content": msg.content, "metadata": msg.metadata}
    print(json.dumps(out, indent=4, default=lambda o: "<not serializable object>"))


def _start_register(redis_config, topic, callback, *args, **kwargs):
    connector = RedisConnector(redis_config)
    connector.register(topics=topic, cb=callback, **kwargs)
    event = threading.Event()
    event.wait()


def channel_monitor_launch():
    """
    Launch a channel monitor for a given channel.
    """
    parser = argparse.ArgumentParser(formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    parser.add_argument(
        "--redis", default="localhost:6379", required=False, help="host:port of redis"
    )
    parser.add_argument(
        "--channel", required=True, help="channel name, e.g. internal/devices/read/samx"
    )
    clargs = parser.parse_args()
    redis_config = clargs.redis
    topic = clargs.channel

    _start_register(redis_config, topic, channel_callback)


def log_callback(msg, log_filter=None):
    """
    Callback for channel monitor.
    """
    msg = msg.value
    print_msg = msg.log_msg["text"]
    if log_filter is not None:
        found = log_filter.lower() in print_msg.lower() or re.search(log_filter, print_msg)
        if found is None:
            return
    print(print_msg)


def log_monitor_launch():
    """
    Launch a log monitor.
    """
    parser = argparse.ArgumentParser(formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    parser.add_argument(
        "--redis", default="localhost:6379", required=False, help="host:port of redis"
    )
    parser.add_argument("--filter", default=None, required=False, help="filter for log messages")
    clargs = parser.parse_args()
    redis_config = clargs.redis
    topic = MessageEndpoints.log()
    log_filter = clargs.filter

    _start_register(redis_config, topic, log_callback, log_filter=log_filter)


# if __name__ == "__main__":
#     import sys

#     sys.argv = ["", "--filter", "sim.py"]
#     log_monitor_launch()



import os

from bec_lib.pdf_writer import PDFWriter


def test_pdf_writer():
    with PDFWriter("test_output.pdf") as file:
        file.write("Hello World")

    assert os.path.exists("test_output.pdf")

    os.remove("test_output.pdf")


def test_pdf_writer_with_title():
    with PDFWriter("test_output.pdf", title="title") as file:
        file.write("Hello World")

    assert os.path.exists("test_output.pdf")

    os.remove("test_output.pdf")



from unittest import mock

import pytest

from bec_lib import MessageEndpoints, messages
from bec_lib.scan_manager import ScanManager


@pytest.fixture
def scan_manager():
    connector = mock.MagicMock()
    manager = ScanManager(connector=connector)
    yield manager
    manager.shutdown()


def test_scan_manager_next_scan_number(scan_manager):
    scan_manager.connector.get.return_value = messages.VariableMessage(value=3)
    assert scan_manager.next_scan_number == 3


def test_scan_manager_next_scan_number_failed(scan_manager):
    scan_manager.connector.get.return_value = None
    assert scan_manager.next_scan_number == -1


def test_scan_manager_next_scan_number_with_int(scan_manager):
    scan_manager.connector.get.return_value = 3
    assert scan_manager.next_scan_number == 3


def test_scan_manager_next_scan_number_setter(scan_manager):
    scan_manager.next_scan_number = 3
    scan_manager.connector.set.assert_called_once_with(
        MessageEndpoints.scan_number(), messages.VariableMessage(value=3)
    )


def test_scan_manager_next_dataset_number(scan_manager):
    scan_manager.connector.get.return_value = messages.VariableMessage(value=3)
    assert scan_manager.next_dataset_number == 3


def test_scan_manager_next_dataset_number_failed(scan_manager):
    scan_manager.connector.get.return_value = None
    assert scan_manager.next_dataset_number == -1


def test_scan_manager_next_dataset_number_with_int(scan_manager):
    scan_manager.connector.get.return_value = 3
    assert scan_manager.next_dataset_number == 3


def test_scan_manager_next_dataset_number_setter(scan_manager):
    scan_manager.next_dataset_number = 3
    scan_manager.connector.set.assert_called_once_with(
        MessageEndpoints.dataset_number(), messages.VariableMessage(value=3)
    )



from __future__ import annotations

import copy
import traceback
from typing import TYPE_CHECKING

from bec_lib import DeviceConfigError, MessageEndpoints, bec_logger, messages

if TYPE_CHECKING:
    from devicemanager import DeviceManagerDS

logger = bec_logger.logger


class ConfigUpdateHandler:
    def __init__(self, device_manager: DeviceManagerDS) -> None:
        self.device_manager = device_manager
        self.connector = self.device_manager.connector
        self.connector.register(
            MessageEndpoints.device_server_config_request(),
            cb=self._device_config_callback,
            parent=self,
        )

    @staticmethod
    def _device_config_callback(msg, *, parent, **_kwargs) -> None:
        logger.info(f"Received request: {msg}")
        parent.parse_config_request(msg.value)

    def parse_config_request(self, msg: messages.DeviceConfigMessage) -> None:
        """Processes a config request. If successful, it emits a config reply

        Args:
            msg (BECMessage.DeviceConfigMessage): Config request

        """
        error_msg = ""
        accepted = True
        try:
            self.device_manager.check_request_validity(msg)
            if msg.content["action"] == "update":
                self._update_config(msg)
            if msg.content["action"] == "add":
                raise NotImplementedError
            if msg.content["action"] == "reload":
                self._reload_config()
                if self.device_manager.failed_devices:
                    msg.metadata["failed_devices"] = self.device_manager.failed_devices

        except DeviceConfigError:
            error_msg = traceback.format_exc()
            accepted = False
        finally:
            self.send_config_request_reply(
                accepted=accepted, error_msg=error_msg, metadata=msg.metadata
            )

    def send_config_request_reply(self, accepted: bool, error_msg: str, metadata: dict) -> None:
        """
        Sends a config request reply

        Args:
            accepted (bool): Whether the request was accepted
            error_msg (str): Error message
            metadata (dict): Metadata of the request
        """
        msg = messages.RequestResponseMessage(
            accepted=accepted, message=error_msg, metadata=metadata
        )
        RID = metadata.get("RID")
        self.device_manager.connector.set(
            MessageEndpoints.device_config_request_response(RID), msg, expire=60
        )

    def _update_config(self, msg: messages.DeviceConfigMessage) -> None:
        for dev, dev_config in msg.content["config"].items():
            device = self.device_manager.devices[dev]
            if "deviceConfig" in dev_config:
                # store old config
                old_config = device._config["deviceConfig"].copy()

                # apply config
                try:
                    self.device_manager.update_config(device.obj, dev_config["deviceConfig"])
                except Exception as exc:
                    self.device_manager.update_config(device.obj, old_config)
                    raise DeviceConfigError(f"Error during object update. {exc}")

                if "limits" in dev_config["deviceConfig"]:
                    limits = {
                        "low": {"value": device.obj.low_limit_travel.get()},
                        "high": {"value": device.obj.high_limit_travel.get()},
                    }
                    self.device_manager.connector.set_and_publish(
                        MessageEndpoints.device_limits(device.name),
                        messages.DeviceMessage(signals=limits),
                    )

            if "enabled" in dev_config:
                device._config["enabled"] = dev_config["enabled"]
                if dev_config["enabled"]:
                    # pylint:disable=protected-access
                    if device.obj._destroyed:
                        self.device_manager.initialize_device(device._config)
                    else:
                        self.device_manager.initialize_enabled_device(device)
                else:
                    self.device_manager.disconnect_device(device.obj)
                    self.device_manager.reset_device(device)

    def _reload_config(self) -> None:
        for _, obj in self.device_manager.devices.items():
            try:
                obj.obj.destroy()
            except Exception:
                logger.warning(f"Failed to destroy {obj.obj.name}")
                raise RuntimeError
        self.device_manager.devices.flush()
        self.device_manager._get_config()
        if self.device_manager.failed_devices:
            self.handle_failed_device_inits()
        return

    def handle_failed_device_inits(self):
        if self.device_manager.failed_devices:
            msg = messages.DeviceConfigMessage(
                action="update",
                config={name: {"enabled": False} for name in self.device_manager.failed_devices},
            )
            self._update_config(msg)
            self.force_update_config_in_redis()
        return

    def force_update_config_in_redis(self):
        config = []
        for name, device in self.device_manager.devices.items():
            device_config = copy.deepcopy(device._config)
            device_config["name"] = name
            config.append(device_config)
        msg = messages.AvailableResourceMessage(resource=config)
        self.device_manager.producer.set(MessageEndpoints.device_config(), msg)



from unittest import mock

import pytest

from bec_lib.device import DeviceBase
from bec_lib.scans import DatasetIdOnHold, HideReport, Metadata, ScanDef, ScanExport, ScanGroup

# pylint: disable=no-member
# pylint: disable=missing-function-docstring
# pylint: disable=redefined-outer-name
# pylint: disable=protected-access


def test_metadata_handler(bec_client_mock):
    client = bec_client_mock
    client.metadata = {"descr": "test", "uid": "12345"}
    with Metadata({"descr": "alignment", "pol": 1}):
        assert client.metadata == {"descr": "alignment", "uid": "12345", "pol": 1}

    assert client.metadata == {"descr": "test", "uid": "12345"}


def test_hide_report_cm(bec_client_mock):
    client = bec_client_mock
    client.scans._hide_report = None
    hrep = HideReport(client.scans)
    with hrep:
        assert client.scans._hide_report is True

    assert client.scans._hide_report is None


def test_dataset_id_on_hold_cm(bec_client_mock):
    client = bec_client_mock
    client.scans._dataset_id_on_hold = None
    dataset_id_on_hold = DatasetIdOnHold(client.scans)
    with mock.patch.object(client, "queue"):
        with dataset_id_on_hold:
            assert client.scans._dataset_id_on_hold is True

    assert client.scans._dataset_id_on_hold is None


def test_dataset_id_on_hold_cm_nested(bec_client_mock):
    client = bec_client_mock
    client.scans._dataset_id_on_hold = None
    dataset_id_on_hold = DatasetIdOnHold(client.scans)
    with mock.patch.object(client, "queue"):
        with dataset_id_on_hold:
            assert client.scans._dataset_id_on_hold is True
            with dataset_id_on_hold:
                assert client.scans._dataset_id_on_hold is True
            assert client.scans._dataset_id_on_hold is True
    assert client.scans._dataset_id_on_hold is None


def test_dataset_id_on_hold_cleanup_on_error(bec_client_mock):
    client = bec_client_mock
    client.scans._dataset_id_on_hold = None
    dataset_id_on_hold = DatasetIdOnHold(client.scans)
    with pytest.raises(AttributeError):
        with mock.patch.object(client, "queue"):
            with dataset_id_on_hold:
                assert client.scans._dataset_id_on_hold is True
                with dataset_id_on_hold:
                    assert client.scans._dataset_id_on_hold is True
                    raise AttributeError()
    assert client.scans._dataset_id_on_hold is None


def test_scan_def_cm(bec_client_mock):
    client = bec_client_mock
    client.scans._scan_def_id = None
    scan_def_id_cm = ScanDef(client.scans)
    with scan_def_id_cm:
        assert isinstance(client.scans._scan_def_id, str)

    assert client.scans._scan_def_id is None


def test_scan_group_cm(bec_client_mock):
    client = bec_client_mock
    client.scans._scan_group = None
    scan_group_cm = ScanGroup(client.scans)
    with scan_group_cm:
        assert isinstance(client.scans._scan_group, str)

    assert client.scans._scan_group is None


@pytest.mark.parametrize("abort_on_ctrl_c", [True, False])
def test_scan_export_cm(abort_on_ctrl_c):
    scan_export = ScanExport("temp")
    scan_export._get_client = mock.MagicMock()
    scan_export._get_client.return_value = mock_client = mock.MagicMock()
    mock_client._service_config = mock_abort = mock.PropertyMock()
    mock_abort.abort_on_ctrl_c = abort_on_ctrl_c
    scan_export._export_to_csv = mock_to_csv = mock.MagicMock()
    if not abort_on_ctrl_c:
        with pytest.raises(RuntimeError):
            with scan_export:
                ...  # Do nothing
    else:
        with scan_export:
            ...  # Do nothgin
        assert mock_to_csv.call_count == 1


def test_parameter_bundler(bec_client_mock):
    client = bec_client_mock
    dev = client.device_manager.devices
    res = client.scans._parameter_bundler((dev.samx, -5, 5, dev.samy, -5, 5), 3)
    assert res == {"samx": [-5, 5], "samy": [-5, 5]}

    res = client.scans._parameter_bundler((dev.samx, -5, 5, 5), 4)
    assert res == {"samx": [-5, 5, 5]}

    res = client.scans._parameter_bundler((-5, 5, 5), 0)
    assert res == (-5, 5, 5)


@pytest.mark.parametrize(
    "in_type,out",
    [
        ("float", (float, int)),
        ("int", int),
        ("list", list),
        ("boolean", bool),
        ("str", str),
        ("dict", dict),
        ("device", DeviceBase),
    ],
)
def test_get_arg_type(bec_client_mock, in_type, out):
    client = bec_client_mock
    res = client.scans.get_arg_type(in_type)
    assert res == out


def test_get_arg_type_raises(bec_client_mock):
    client = bec_client_mock
    with pytest.raises(TypeError):
        client.scans.get_arg_type("not_existing")



import argparse

import yaml

from bec_lib import messages
from bec_lib.endpoints import MessageEndpoints
from bec_lib.redis_connector import RedisConnector

parser = argparse.ArgumentParser(formatter_class=argparse.ArgumentDefaultsHelpFormatter)
parser.add_argument(
    "--config", default="./bec_lib/bec_lib/configs/demo_config.yaml", help="path to the config file"
)
parser.add_argument("--redis", default="localhost:6379", help="redis host and port")

clargs = parser.parse_args()
connector = RedisConnector(clargs.redis)

with open(clargs.config, "r", encoding="utf-8") as stream:
    data = yaml.safe_load(stream)
for name, device in data.items():
    device["name"] = name
config_data = list(data.values())
msg = messages.AvailableResourceMessage(resource=config_data)
connector.set(MessageEndpoints.device_config(), msg)



import copy
import functools
import os
from unittest import mock

import numpy as np
import pytest
import yaml

import bec_lib
from bec_lib import MessageEndpoints, messages
from bec_server.device_server.devices.devicemanager import DeviceManagerDS

# pylint: disable=missing-function-docstring
# pylint: disable=protected-access


class ControllerMock:
    def __init__(self, parent) -> None:
        self.parent = parent

    def on(self):
        self.parent._connected = True

    def off(self):
        self.parent._connected = False


class DeviceMock:
    def __init__(self) -> None:
        self._connected = False
        self.name = "name"

    @property
    def connected(self):
        return self._connected


class DeviceControllerMock(DeviceMock):
    def __init__(self) -> None:
        super().__init__()
        self.controller = ControllerMock(self)


class EpicsDeviceMock(DeviceMock):
    def wait_for_connection(self, timeout):
        self._connected = True


@pytest.mark.parametrize("device_manager_class", [DeviceManagerDS])
def test_device_init(dm_with_devices):
    device_manager = dm_with_devices
    for dev in device_manager.devices.values():
        if not dev.enabled:
            continue
        assert dev.initialized is True


@pytest.mark.parametrize("device_manager_class", [DeviceManagerDS])
def test_device_proxy_init(dm_with_devices):
    device_manager = dm_with_devices
    assert "sim_proxy_test" in device_manager.devices.keys()
    assert "proxy_cam_test" in device_manager.devices.keys()
    assert "image" in device_manager.devices["proxy_cam_test"].obj.registered_proxies.values()
    assert (
        "sim_proxy_test" in device_manager.devices["proxy_cam_test"].obj.registered_proxies.keys()
    )


@pytest.mark.parametrize(
    "obj,raises_error",
    [(DeviceMock(), True), (DeviceControllerMock(), False), (EpicsDeviceMock(), False)],
)
@pytest.mark.parametrize("device_manager_class", [DeviceManagerDS])
def test_connect_device(dm_with_devices, obj, raises_error):
    device_manager = dm_with_devices
    if raises_error:
        with pytest.raises(ConnectionError):
            device_manager.connect_device(obj)
        return
    device_manager.connect_device(obj)


@pytest.mark.parametrize("device_manager_class", [DeviceManagerDS])
def test_disable_unreachable_devices(device_manager, session_from_test_config):
    def get_config_from_mock():
        device_manager._session = copy.deepcopy(session_from_test_config)
        device_manager._load_session()

    def mocked_failed_connection(obj):
        if obj.name == "samx":
            raise ConnectionError

    config_reply = messages.RequestResponseMessage(accepted=True, message="")

    with mock.patch.object(device_manager, "connect_device", wraps=mocked_failed_connection):
        with mock.patch.object(device_manager, "_get_config", get_config_from_mock):
            with mock.patch.object(
                device_manager.config_helper, "wait_for_config_reply", return_value=config_reply
            ):
                with mock.patch.object(device_manager.config_helper, "wait_for_service_response"):
                    device_manager.initialize("")
                    assert device_manager.config_update_handler is not None
                    assert device_manager.devices.samx.enabled is False
                    msg = messages.DeviceConfigMessage(
                        action="update", config={"samx": {"enabled": False}}
                    )


@pytest.mark.parametrize("device_manager_class", [DeviceManagerDS])
def test_flyer_event_callback(dm_with_devices):
    device_manager = dm_with_devices
    samx = device_manager.devices.samx
    samx.metadata = {"scan_id": "12345"}

    device_manager._obj_flyer_callback(
        obj=samx.obj, value={"data": {"idata": np.random.rand(20), "edata": np.random.rand(20)}}
    )
    pipe = device_manager.connector.pipeline()
    bundle, progress = pipe._pipe_buffer[-2:]

    # check connector method
    assert bundle[0] == "send"
    assert progress[0] == "set_and_publish"

    # check endpoint
    assert bundle[1][0] == MessageEndpoints.device_read("samx").endpoint
    assert progress[1][0] == MessageEndpoints.device_progress("samx").endpoint

    # check message
    bundle_msg = bundle[1][1]
    assert len(bundle_msg) == 20

    progress_msg = progress[1][1]
    assert progress_msg.content["status"] == 20


@pytest.mark.parametrize("device_manager_class", [DeviceManagerDS])
def test_obj_progress_callback(dm_with_devices):
    device_manager = dm_with_devices
    samx = device_manager.devices.samx
    samx.metadata = {"scan_id": "12345"}

    with mock.patch.object(device_manager, "connector") as mock_connector:
        device_manager._obj_progress_callback(obj=samx.obj, value=1, max_value=2, done=False)
        mock_connector.set_and_publish.assert_called_once_with(
            MessageEndpoints.device_progress("samx"),
            messages.ProgressMessage(
                value=1, max_value=2, done=False, metadata={"scan_id": "12345"}
            ),
        )


@pytest.mark.parametrize(
    "value", [np.empty(shape=(10, 10)), np.empty(shape=(100, 100)), np.empty(shape=(1000, 1000))]
)
@pytest.mark.parametrize("device_manager_class", [DeviceManagerDS])
def test_obj_monitor_callback(dm_with_devices, value):
    device_manager = dm_with_devices
    eiger = device_manager.devices.eiger
    eiger.metadata = {"scan_id": "12345"}
    value_size = len(value.tobytes()) / 1e6  # MB
    max_size = 1000
    with mock.patch.object(device_manager, "connector") as mock_connector:
        device_manager._obj_callback_monitor(obj=eiger.obj, value=value)
        mock_connector.xadd.assert_called_once_with(
            MessageEndpoints.device_monitor(eiger.name),
            {
                "data": messages.DeviceMonitorMessage(
                    device=eiger.name, data=value, metadata={"scan_id": "12345"}
                )
            },
            max_size=int(min(100, max_size / value_size)),
        )



from math import inf

import lmfit

from bec_lib.lmfit_serializer import deserialize_param_object, serialize_lmfit_params


def test_serialize_lmfit_params():
    params = lmfit.Parameters()
    params.add("a", value=1.0, vary=True)
    params.add("b", value=2.0, vary=False)
    result = serialize_lmfit_params(params)
    assert result == {
        "a": {
            "name": "a",
            "value": 1.0,
            "vary": True,
            "min": -inf,
            "max": inf,
            "expr": None,
            "brute_step": None,
        },
        "b": {
            "name": "b",
            "value": 2.0,
            "vary": False,
            "min": -inf,
            "max": inf,
            "expr": None,
            "brute_step": None,
        },
    }

    obj = deserialize_param_object(result)
    assert obj == params



"""
This module contains the ScanItem and ScanStorage classes. The ScanItem class is used to store
information about a scan. The ScanStorage class is used to store scan items.
"""

from __future__ import annotations

import builtins
import datetime
import sys
import threading
import time
from collections import defaultdict, deque
from typing import TYPE_CHECKING

from bec_lib import messages
from bec_lib.async_data import AsyncDataHandler
from bec_lib.logger import bec_logger
from bec_lib.scan_data import ScanData
from bec_lib.utils import threadlocked

if TYPE_CHECKING:
    from bec_lib.scan_manager import ScanManager

logger = bec_logger.logger

try:
    import pandas as pd
except ImportError:
    logger.info("Unable to import `pandas` optional dependency")


class ScanItem:
    status: dict

    # pylint: disable=too-many-arguments
    def __init__(
        self,
        scan_manager: ScanManager,
        queue_id: str,
        scan_number: list,
        scan_id: list,
        status: str,
        **_kwargs,
    ) -> None:
        self.scan_manager = scan_manager
        self._queue_id = queue_id
        self.scan_number = scan_number
        self.scan_id = scan_id
        self.status = status
        self.data = ScanData()
        self.async_data = {}
        self.baseline = ScanData()
        self._async_data_handler = AsyncDataHandler(scan_manager.connector)
        self.open_scan_defs = set()
        self.open_queue_group = None
        self.num_points = None
        self.start_time = None
        self.end_time = None
        self.scan_report_instructions = []
        self.callbacks = []
        self.bec = builtins.__dict__.get("bec")

    @property
    def queue(self):
        """get the queue item for the current scan item"""
        return self.scan_manager.queue_storage.find_queue_item_by_ID(self._queue_id)

    def emit_data(self, scan_msg: messages.ScanMessage) -> None:
        self.bec.callbacks.run("scan_segment", scan_msg.content, scan_msg.metadata)
        self._run_request_callbacks("scan_segment", scan_msg.content, scan_msg.metadata)

    def emit_status(self, scan_status: messages.ScanStatusMessage) -> None:
        self.bec.callbacks.run("scan_status", scan_status.content, scan_status.metadata)
        self._run_request_callbacks("scan_status", scan_status.content, scan_status.metadata)

    def _run_request_callbacks(self, event_type: str, data: dict, metadata: dict):
        for rid in self.queue.requestIDs:
            req = self.scan_manager.request_storage.find_request_by_ID(rid)
            if req is None:
                continue
            req.callbacks.run(event_type, data, metadata)

    def poll_callbacks(self):
        for rid in self.queue.requestIDs:
            req = self.scan_manager.request_storage.find_request_by_ID(rid)
            if req is None:
                continue
            req.callbacks.poll()

    def to_pandas(self) -> pd.DataFrame:
        """convert to pandas dataframe"""
        if "pandas" not in sys.modules:
            raise ImportError("Install `pandas` to use to_pandas() method")

        tmp = defaultdict(list)
        for scan_msg in self.data.messages.values():
            scan_msg_data = scan_msg.content["data"]
            for dev, dev_data in scan_msg_data.items():
                for signal, signal_data in dev_data.items():
                    for key, value in signal_data.items():
                        tmp[(dev, signal, key)].append(value)
        return pd.DataFrame(tmp)

    def _update_async_data(self):
        self.async_data = self._async_data_handler.get_async_data_for_scan(self.scan_id)
        # msg = messages.ScanMessage(point_id=0, scan_id=self.scan_id, data=data)
        # self.async_data.set(0, msg)

    def __eq__(self, other):
        return self.scan_id == other.scan_id

    def describe(self) -> str:
        """describe the scan item"""
        start_time = (
            f"\tStart time: {datetime.datetime.fromtimestamp(self.start_time).strftime('%c')}\n"
            if self.start_time
            else ""
        )
        end_time = (
            f"\tEnd time: {datetime.datetime.fromtimestamp(self.end_time).strftime('%c')}\n"
            if self.end_time
            else ""
        )
        elapsed_time = (
            f"\tElapsed time: {(self.end_time-self.start_time):.1f} s\n"
            if self.end_time and self.start_time
            else ""
        )
        scan_id = f"\tScan ID: {self.scan_id}\n" if self.scan_id else ""
        scan_number = f"\tScan number: {self.scan_number}\n" if self.scan_number else ""
        num_points = f"\tNumber of points: {self.num_points}\n" if self.num_points else ""
        details = start_time + end_time + elapsed_time + scan_id + scan_number + num_points
        return details

    def __str__(self) -> str:
        return f"ScanItem:\n {self.describe()}"


class ScanStorage:
    """stores scan items"""

    def __init__(self, scan_manager: ScanManager, maxlen=50, init_scan_number=0) -> None:
        self.scan_manager = scan_manager
        self.storage = deque(maxlen=maxlen)
        self.last_scan_number = init_scan_number
        self._lock = threading.RLock()

    @property
    def current_scan_info(self) -> dict:
        """get the current scan info from the scan queue"""
        scan_queue = self.scan_manager.queue_storage.current_scan_queue
        if not scan_queue or not scan_queue["primary"].get("info"):
            return None
        return scan_queue["primary"].get("info")[0]

    @property
    def current_scan(self) -> ScanItem | None:
        """get the current scan item"""
        if not self.current_scan_id:
            return None
        return self.find_scan_by_ID(scan_id=self.current_scan_id[0])

    @property
    def current_scan_id(self) -> str | None:
        """get the current scan_id"""
        if self.current_scan_info is None:
            return None
        return self.current_scan_info.get("scan_id")

    @threadlocked
    def find_scan_by_ID(self, scan_id: str) -> ScanItem | None:
        """find a scan item based on its scan_id"""
        for scan in self.storage:
            if scan_id == scan.scan_id:
                return scan
        return None

    def update_with_scan_status(self, scan_status: messages.ScanStatusMessage) -> None:
        """update scan item in storage with a new ScanStatusMessage"""

        scan_id = scan_status.content["scan_id"]
        if not scan_id:
            return

        scan_number = scan_status.content["info"].get("scan_number")
        if scan_number:
            self.last_scan_number = scan_number

        while True:
            scan_item = self.find_scan_by_ID(scan_id=scan_status.content["scan_id"])
            if scan_item:
                break
            time.sleep(0.1)

        # update timestamps
        if scan_status.content.get("status") == "open":
            scan_item.start_time = scan_status.content.get("timestamp")
        elif scan_status.content.get("status") == "closed":
            scan_item.end_time = scan_status.content.get("timestamp")

        # update status message
        scan_item.status = scan_status.content.get("status")
        scan_item.status_message = scan_status

        # update total number of points
        if scan_status.content["info"].get("num_points"):
            scan_item.num_points = scan_status.content["info"].get("num_points")

        # update scan number
        if scan_item.scan_number is None:
            scan_item.scan_number = scan_number

        # add scan report info
        scan_item.scan_report_instructions = scan_status.content["info"].get(
            "scan_report_instructions"
        )

        # add scan def id
        scan_def_id = scan_status.content["info"].get("scan_def_id")
        if scan_def_id:
            if scan_status.content.get("status") != "open":
                scan_item.open_scan_defs.remove(scan_def_id)
            else:
                scan_item.open_scan_defs.add(scan_def_id)

        # add queue group
        scan_item.open_queue_group = scan_status.content["info"].get("queue_group")

        # update async data
        if scan_status.content.get("status") != "open":
            # pylint: disable=protected-access
            scan_item._update_async_data()

        # run status callbacks
        scan_item.emit_status(scan_status)

    def add_scan_segment(self, scan_msg: messages.ScanMessage) -> None:
        """update a scan item with a new scan segment"""
        logger.info(
            f"Received scan segment {scan_msg.content['point_id']} for scan"
            f" {scan_msg.metadata['scan_id']}: "
        )
        while True:
            with self._lock:
                for scan_item in self.storage:
                    if scan_item.scan_id == scan_msg.metadata["scan_id"]:
                        scan_item.data.set(scan_msg.content["point_id"], scan_msg)
                        scan_item.emit_data(scan_msg)
                        return
            time.sleep(0.01)

    def add_scan_baseline(self, scan_msg: messages.ScanBaselineMessage) -> None:
        """update a scan item with a new scan baseline"""
        logger.info(f"Received scan baseline for scan {scan_msg.metadata['scan_id']}: ")
        while True:
            with self._lock:
                for scan_item in self.storage:
                    if scan_item.scan_id == scan_msg.metadata["scan_id"]:
                        point = len(scan_item.baseline)
                        scan_item.baseline.set(point, scan_msg)
                        return
            time.sleep(0.01)

    @threadlocked
    def add_scan_item(self, queue_id: str, scan_number: list, scan_id: list, status: str):
        """append new scan item to scan storage"""
        self.storage.append(ScanItem(self.scan_manager, queue_id, scan_number, scan_id, status))

    @threadlocked
    def update_with_queue_status(self, queue_msg: messages.ScanQueueStatusMessage):
        """create new scan items based on their existence in the queue info"""
        queue_info = queue_msg.content["queue"]["primary"].get("info")
        for queue_item in queue_info:
            # append = True
            # for scan_obj in self.storage:
            #     if len(set(scan_obj.scan_id) & set(queue_item["scan_id"])) > 0:
            #         append = False
            if not any(queue_item["is_scan"]):
                continue

            for ii, scan in enumerate(queue_item["scan_id"]):
                if self.find_scan_by_ID(scan):
                    continue

                logger.debug(f"Appending new scan: {queue_item}")
                self.add_scan_item(
                    queue_id=queue_item["queue_id"],
                    scan_number=queue_item["scan_number"][ii],
                    scan_id=queue_item["scan_id"][ii],
                    status=queue_item["status"],
                )



from unittest import mock

import pytest

from bec_lib import messages
from bec_lib.channel_monitor import channel_callback, channel_monitor_launch, log_callback
from bec_lib.redis_connector import MessageObject


def test_channel_monitor_callback():
    with mock.patch("builtins.print") as mock_print:
        msg = messages.DeviceMessage(
            signals={"x": {"value": 1}, "y": {"value": 2}}, metadata={"name": "test"}
        )
        msg_obj = MessageObject("test", msg)
        channel_callback(msg_obj)
        mock_print.assert_called_once()


def test_channel_monitor_start_register():
    with mock.patch("bec_lib.channel_monitor.argparse") as mock_argparse:
        with mock.patch("bec_lib.channel_monitor.RedisConnector") as mock_connector:
            with mock.patch("bec_lib.channel_monitor.threading") as mock_threading:
                clargs = mock.MagicMock()
                mock_argparse.ArgumentParser().parse_args.return_value = clargs
                clargs.config = "test_config"
                clargs.channel = "test_channel"
                mock_threading.Event().wait.return_value = True
                mock_connector.return_value = mock.MagicMock()
                channel_monitor_launch()
                mock_connector().register.assert_called_once()
                mock_threading.Event().wait.assert_called_once()


def test_log_monitor_callback_without_filter():
    with mock.patch("builtins.print") as mock_print:
        msg = messages.LogMessage(log_type="info", log_msg={"text": "test"})
        msg_obj = MessageObject("test", msg)
        log_callback(msg_obj)
        mock_print.assert_called_once_with("test")


@pytest.mark.parametrize(
    "text, log_filter, printed",
    [
        ("test", "te", True),
        ("test", "no", False),
        ("test", None, True),
        ("test", ".*", True),
        ("test", ".*no.*", False),
    ],
)
def test_log_monitor_callback_filter(text, log_filter, printed):
    with mock.patch("builtins.print") as mock_print:
        msg = messages.LogMessage(log_type="info", log_msg={"text": text})
        msg_obj = MessageObject("test", msg)
        log_callback(msg_obj, log_filter=log_filter)
        if printed:
            mock_print.assert_called_once_with("test")
        else:
            mock_print.assert_not_called()



import pytest
from ophyd import Component as Cpt
from ophyd import Device, Signal

from bec_lib.bec_errors import DeviceConfigError
from bec_server.device_server.devices.device_serializer import get_device_info


class DummyDeviceWithConflictingSignalNames(Device):
    # This device has a signal with the same name as a protected method
    # in the Device class
    enabled = Cpt(Signal, value=0)


class DummyDeviceWithConflictingName(Device):
    """This device will be assigned a protected name"""


class DummyDeviceWithConflictingUserAccess(Device):
    """This device will be assigned a protected name"""

    USER_ACCESS = ["enabled"]

    def enabled(self):
        pass


class DummyDeviceWithConflictingSubDevice(Device):
    """This device will be assigned a protected name"""

    sub_device = Cpt(DummyDeviceWithConflictingSignalNames)


@pytest.mark.parametrize(
    "obj",
    [
        DummyDeviceWithConflictingSignalNames(name="test"),
        DummyDeviceWithConflictingSignalNames(name="enabled"),
        DummyDeviceWithConflictingSubDevice(name="test"),
        DummyDeviceWithConflictingUserAccess(name="test"),
    ],
)
def test_get_device_info(obj):
    with pytest.raises(DeviceConfigError):
        _ = get_device_info(obj)



from unittest import mock

import lmfit
import numpy as np
import pytest

from bec_server.data_processing.dap_service import DAPError
from bec_server.data_processing.lmfit1d_service import LmfitService1D


@pytest.fixture
def lmfit_service():
    yield LmfitService1D(model="GaussianModel", continuous=False, client=mock.MagicMock())


@pytest.mark.parametrize("model, exists", [("GaussianModel", True), ("ModelDoesntExist", False)])
def test_LmfitService1D(model, exists):
    client = mock.MagicMock()
    if exists:
        service = LmfitService1D(model=model, client=client)
        return
    with pytest.raises(AttributeError):
        service = LmfitService1D(model=model, client=client)


def test_LmfitService1D_available_models(lmfit_service):
    models = lmfit_service.available_models()
    assert len(models) > 0
    assert all(issubclass(model, lmfit.model.Model) for model in models)
    assert all(
        model.__name__ not in ["Gaussian2dModel", "ExpressionModel", "Model", "SplineModel"]
        for model in models
    )


def test_LmfitService1D_get_provided_services(lmfit_service):
    services = lmfit_service.get_provided_services()
    assert isinstance(services, dict)
    assert len(services) > 0
    for model, service in services.items():
        assert isinstance(service, dict)
        assert "class" in service
        assert "user_friendly_name" in service
        assert "class_doc" in service
        assert "run_doc" in service


def test_LmfitService1D_get_data_from_current_scan_without_devices(lmfit_service):
    scan_item = mock.MagicMock()
    scan_item.data = mock.MagicMock()
    scan_item.data[0].metadata = {"scan_report_devices": ["device_x", "device_y"]}

    data = lmfit_service.get_data_from_current_scan(scan_item)
    assert data is None


def test_LmfitService1D_get_data_from_current_scan(lmfit_service):
    scan_item = mock.MagicMock()
    scan_item.data = mock.MagicMock()
    lmfit_service.device_x = "device_x"
    lmfit_service.signal_x = "signal_x"
    lmfit_service.device_y = "device_y"
    lmfit_service.signal_y = "signal_y"

    scan_item.data = {
        "device_x": {"signal_x": {"value": [1, 2, 3], "timestamp": 0}},
        "device_y": {"signal_y": {"value": [4, 5, 6], "timestamp": 0}},
    }
    data = lmfit_service.get_data_from_current_scan(scan_item)
    assert all(data["x"] == np.array([1, 2, 3]))
    assert all(data["y"] == np.array([4, 5, 6]))


@pytest.mark.parametrize(
    "scan_data",
    [
        {
            "device_x": {"signal_x": {"value": [1, 2], "timestamp": 0}},
            "device_y": {"signal_y": {"value": [4, 5], "timestamp": 0}},
        },
        {"device_y": {"signal_y": {"value": [4, 5, 6], "timestamp": 0}}},
        {"device_x": {"signal_x": {"value": [1, 2], "timestamp": 0}}},
        None,
    ],
)
def test_LmfitService1D_get_data_from_current_scan_returns_None(lmfit_service, scan_data):
    scan_item = mock.MagicMock()
    scan_item.data = mock.MagicMock()
    lmfit_service.device_x = "device_x"
    lmfit_service.signal_x = "signal_x"
    lmfit_service.device_y = "device_y"
    lmfit_service.signal_y = "signal_y"
    scan_item.data = scan_data
    data = lmfit_service.get_data_from_current_scan(scan_item)
    assert data is None


def test_LmfitService1D_process(lmfit_service):
    lmfit_service.data = {"x": [1, 2, 3], "y": [4, 5, 6], "x_lim": False, "x_original": [1, 2, 3]}
    lmfit_service.model = mock.MagicMock()

    result = lmfit_service.process()
    assert isinstance(result, tuple)
    assert isinstance(result[0], dict)
    assert isinstance(result[1], dict)
    lmfit_service.model.fit.assert_called_once_with([4, 5, 6], x=[1, 2, 3])


def test_LmfitService1D_on_scan_status_update(lmfit_service):
    with mock.patch.object(lmfit_service, "process_until_finished") as process_until_finished:
        lmfit_service.on_scan_status_update({"status": "running"}, {})
        process_until_finished.assert_called_once()


def test_LmfitService1D_on_scan_status_update_finishes_on_closed_scans(lmfit_service):
    with mock.patch("bec_server.data_processing.lmfit1d_service.threading") as threading:
        event = threading.Event()
        lmfit_service.finish_event = event
        with mock.patch.object(lmfit_service, "process_until_finished") as process_until_finished:
            lmfit_service.on_scan_status_update({"status": "closed"}, {})
            event.set.assert_called_once()
            process_until_finished.assert_not_called()
            assert lmfit_service.finish_event is None


def test_LmfitService1D_process_until_finished(lmfit_service):
    event = mock.MagicMock()
    event.is_set.side_effect = [False, False, True]

    with mock.patch.object(lmfit_service, "get_data_from_current_scan") as get_data:
        get_data.return_value = {"x": [1, 2, 3], "y": [4, 5, 6]}
        with mock.patch.object(lmfit_service, "process") as process:
            process.return_value = ({"result": "result"}, {"metadata": "metadata"})
            lmfit_service.process_until_finished(event)
            assert get_data.call_count == 2
            assert process.call_count == 2
            assert lmfit_service.client.connector.xadd.call_count == 2


def test_LmfitService1D_configure(lmfit_service):
    with pytest.raises(DAPError):
        lmfit_service.configure()


def test_LmfitService1D_configure_selected_devices(lmfit_service):
    lmfit_service.continuous = False
    with pytest.raises(DAPError):
        lmfit_service.configure(selected_device=["bpm4i", "bpm4i"])

    with mock.patch.object(lmfit_service, "get_data_from_current_scan") as get_data:
        get_data.return_value = {"x": [1, 2, 3], "y": [4, 5, 6]}
        lmfit_service.configure(
            selected_device=["bpm4i", "bpm4i"], device_x="samx", signal_x="samx"
        )
        get_data.assert_called_once()


def test_LmfitService1D_get_model(lmfit_service):
    model = lmfit_service.get_model("GaussianModel")
    assert model.__name__ == "GaussianModel"
    assert issubclass(model, lmfit.model.Model)

    with pytest.raises(ValueError):
        lmfit_service.get_model("ModelDoesntExist")



"""
This module contains functions for serializing and deserializing lmfit objects.
"""

from lmfit import Parameter, Parameters


def serialize_param_object(param: Parameter) -> dict:
    """
    Serialize lmfit.Parameter object to JSON-serializable dictionary.

    Args:
        param (Parameter): Parameter object

    Returns:
        dict: Dictionary representation of the parameter
    """
    obj = {
        "name": param.name,
        "value": param.value,
        "vary": param.vary,
        "min": param.min,
        "max": param.max,
        "expr": param.expr,
        "brute_step": param.brute_step,
    }
    return obj


def serialize_lmfit_params(params: Parameters) -> dict:
    """
    Serialize lmfit.Parameters object to JSON-serializable dictionary.

    Args:
        params (Parameters): Parameters object containing lmfit.Parameter objects

    Returns:
        dict: Dictionary representation of the parameters
    """
    if not params:
        return {}
    if isinstance(params, Parameters):
        return {k: serialize_param_object(v) for k, v in params.items()}
    if isinstance(params, list):
        return {v.name: serialize_param_object(v) for v in params}


def deserialize_param_object(obj: dict) -> Parameter:
    """
    Deserialize dictionary representation of lmfit.Parameter object.

    Args:
        obj (dict): Dictionary representation of the parameters

    Returns:
        Parameter: Parameter object
    """
    param = Parameters()
    for k, v in obj.items():
        v.pop("name")
        param.add(k, **v)
    return param
